import logging
import subprocess
import sys

import typer

from cmapi_server.logging_management import dict_config, add_logging_level, enable_console_logging
from mcs_cluster_tool import (
    cluster_app, cmapi_app, backup_commands, restore_commands, tools_commands
)
from mcs_cluster_tool.constants import MCS_CLI_LOG_CONF_PATH


# don't show --install-completion and --show-completion options in help message
app = typer.Typer(
    add_completion=False,
    help=(
        'The  MCS  Command  Line  Interface is a unified tool to manage your '
        'MCS services'
    ),
    rich_markup_mode='rich',
)
app.add_typer(cluster_app.app)
# keep this only for potential backward compatibility and userfriendliness
app.add_typer(cluster_app.app, name='cluster', hidden=True)
app.add_typer(cmapi_app.app, name='cmapi')
app.command(
    'backup', rich_help_panel='Tools commands'
)(backup_commands.backup)
app.command(
    'dbrm_backup', rich_help_panel='Tools commands'
)(backup_commands.dbrm_backup)
app.command(
    'restore', rich_help_panel='Tools commands'
)(restore_commands.restore)
app.command(
    'dbrm_restore', rich_help_panel='Tools commands'
)(restore_commands.dbrm_restore)
app.command(
    'cskeys', rich_help_panel='Tools commands',
    short_help=(
        'Generate a random AES encryption key and init vector and write '
        'them to disk.'
    )
)(tools_commands.cskeys)
app.command(
    'cspasswd', rich_help_panel='Tools commands',
    short_help='Encrypt a Columnstore plaintext password.'
)(tools_commands.cspasswd)
app.command(
    'bootstrap-single-node', rich_help_panel='Tools commands',
)(tools_commands.bootstrap_single_node)
app.command(
    'review', rich_help_panel='Tools commands',
    short_help=(
        'Provides useful functions to review and troubleshoot the MCS cluster.'
    )
)(tools_commands.review)
app.add_typer(
    tools_commands.sentry_app, name='sentry', rich_help_panel='Tools commands', hidden=True
)
app.command(
    'install_es', rich_help_panel='Tools commands',
)(tools_commands.install_es)


@app.command(
        name='help-all', help='Show help for all commands in man page style.',
        add_help_option=False
)
def help_all():
    # Open the man page in interactive mode
    subprocess.run(['man', 'mcs'])


@app.callback()
def main(verbose: bool = typer.Option(False, '--verbose', '-v', help='Enable verbose logging to console')):
    '''Add a -v option and setup logging in every subcommand'''
    setup_logging(verbose)


def setup_logging(verbose: bool = False) -> None:
    add_logging_level('TRACE', 5)
    dict_config(MCS_CLI_LOG_CONF_PATH)
    if verbose:
        for logger_name in ("", "mcs_cli"):
            current_logger = logging.getLogger(logger_name)
            current_logger.setLevel(logging.DEBUG)
            enable_console_logging(current_logger)


if __name__ == '__main__':
    logger = logging.getLogger('mcs_cli')
    # add separator between cli commands logging
    logger.debug(f'{"-":-^80}')
    cl_args_line = ' '.join(sys.argv[1:])
    logger.debug(f'Called "mcs {cl_args_line}"')
    app(prog_name='mcs')



import os
from collections.abc import Generator
from contextlib import contextmanager


@contextmanager
def change_directory(new_dir: str) -> Generator[None, None, None]:
    """
    Context manager for temporarily changing the current working directory.

    :param new_dir: Directory to change to
    """
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(old_dir)


def drop_timestamp(data: dict) -> dict:
    """
    Drop the timestamp from the data dictionary.
    """
    if "timestamp" in data:
        del data["timestamp"]
    return data


def assert_dict_includes(this: dict, other: dict) -> None:
    """
    Check if the current dictionary includes all keys and values from another dictionary.
    """
    for key, value in other.items():
        assert key in this
        assert this[key] == value, f"Key '{key}' does not match: {this[key]} != {value}"



"""
This script tests concurrent inserts on a given table.
Example Usage (in Mysql Test Framework):

  CREATE TABLE t1 (a INT) ENGINE=rocksdb;

  let $exec = python suite/rocksdb/t/rocksdb_concurrent_insert.py \
                     root 127.0.0.1 $MASTER_MYPORT test t1 100 4;
  exec $exec;

"""
import cStringIO
import hashlib
import MySQLdb
import os
import random
import signal
import sys
import threading
import time
import string

def get_insert(table_name, idx):
  return """INSERT INTO %s (a) VALUES (%d)""" % (table_name, idx)

class Inserter(threading.Thread):
  Instance = None
  def __init__(self, con, table_name, num_inserts):
    threading.Thread.__init__(self)
    self.finished = False
    self.num_inserts = num_inserts
    con.autocommit(False)
    self.con = con
    self.rand = random.Random()
    self.exception = None
    self.table_name = table_name
    Inserter.Instance = self
    self.start()
  def run(self):
    try:
      self.runme()
    except Exception, e:
      self.exception = traceback.format_exc()
      print "caught (%s)" % e
    finally:
      self.finish()
  def runme(self):
    cur = self.con.cursor()
    for i in xrange(self.num_inserts):
      try:
        cur.execute(get_insert(self.table_name, i))
        r = self.rand.randint(1,10)
        if r < 4:
          self.con.commit()
      except:
        cur = self.con.cursor()
    try:
      self.con.commit()
    except Exception, e:
      self.exception = traceback.format_exc()
      print "caught (%s)" % e
      pass
  def finish(self):
    self.finished = True

if __name__ == '__main__':
  if len(sys.argv) != 8:
    print "Usage: rocksdb_concurrent_insert.py user host port db_name " \
          "table_name num_inserts num_threads"
    sys.exit(1)

  user = sys.argv[1]
  host = sys.argv[2]
  port = int(sys.argv[3])
  db = sys.argv[4]
  table_name = sys.argv[5]
  num_inserts = int(sys.argv[6])
  num_workers = int(sys.argv[7])

  worker_failed = False
  workers = []
  for i in xrange(num_workers):
    inserter = Inserter(
      MySQLdb.connect(user=user, host=host, port=port, db=db), table_name,
      num_inserts)
    workers.append(inserter)

  for w in workers:
    w.join()
    if w.exception:
      print "Worker hit an exception:\n%s\n" % w.exception
      worker_failed = True

  if worker_failed:
    sys.exit(1)



import logging
import socket
import unittest
from unittest.mock import patch

from lxml import etree
from mcs_node_control.models.node_config import NodeConfig

from cmapi_server import node_manipulation
from cmapi_server.managers.application import (
    AppStatefulConfig, StatefulConfigModel, StatefulFlagsModel
)
from cmapi_server.constants import MCS_DATA_PATH
from cmapi_server.exceptions import CMAPIBasicError
from cmapi_server.node_manipulation import (
    add_dbroots_of_other_nodes,
    remove_dbroots_of_node,
    update_dbroots_of_read_replicas,
)
from cmapi_server.test.mock_resolution import (
    MockResolutionBuilder,
    make_local_resolution_builder,
)
from cmapi_server.test.unittest_global import BaseNodeManipTestCase, tmp_mcs_config_filename, single_node_xml

logging.basicConfig(level='DEBUG')

class NodeManipTester(BaseNodeManipTestCase):
    LOCAL_NODE = 'local.node'
    REMOTE_NODE = 'remote.node'
    LOCAL_IP = '104.17.191.14'
    REMOTE_IP = '203.0.113.5'

    def test_add_remove_node(self):
        # Toggle shared storage as requested and remember to restore it back.
        original_shared_storage = self._set_shared_storage(True)
        self.tmp_files = (
            './test-output0.xml', './test-output1.xml', './test-output2.xml'
        )

        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ), patch(
            'cmapi_server.node_manipulation.update_dbroots_of_read_replicas'
        ) as mock_update_dbroots_of_read_replicas:
            node_manipulation.add_node(
                self.LOCAL_NODE, tmp_mcs_config_filename, self.tmp_files[0]
            )
            mock_update_dbroots_of_read_replicas.assert_not_called()

            node_manipulation.add_node(
                self.REMOTE_IP, self.tmp_files[0], self.tmp_files[1]
            )
            # No read replicas present
            mock_update_dbroots_of_read_replicas.assert_not_called()

        # get a NodeConfig, read test.xml
        # look for some of the expected changes.
        # Total verification will take too long to code up right now.
        nc = NodeConfig()
        root = nc.get_current_config_root(self.tmp_files[1])
        pms_node_ipaddr = root.find('./PMS1/IPAddr')
        self.assertEqual(pms_node_ipaddr.text, self.LOCAL_IP)
        pms_node_ipaddr = root.find('./PMS2/IPAddr')
        self.assertEqual(pms_node_ipaddr.text, self.REMOTE_IP)
        node = root.find('./ExeMgr2/IPAddr')
        self.assertEqual(node.text, self.REMOTE_IP)

        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ), patch(
            'cmapi_server.node_manipulation.update_dbroots_of_read_replicas'
        ) as mock_update_dbroots_of_read_replicas:
            node_manipulation.remove_node(
                self.LOCAL_NODE, self.tmp_files[1], self.tmp_files[2], test_mode=True
            )
            # No read replicas present
            mock_update_dbroots_of_read_replicas.assert_not_called()

        nc = NodeConfig()
        root = nc.get_current_config_root(self.tmp_files[2])
        node = root.find('./PMS1/IPAddr')
        self.assertEqual(node.text, self.REMOTE_IP)
        # TODO: Fix node_manipulation add_node logic and _replace_localhost
        # node = root.find('./PMS2/IPAddr')
        # self.assertEqual(node, None)
        _ = self._set_shared_storage(original_shared_storage)

    def test_add_remove_read_replica_node(self):
        """add_node(read_replica=True) should add a read replica node into the config,
        it does not add a WriteEngineServer (WES) and does not own dbroots"""
        self.tmp_files = (
            './config_output_rw.xml', './config_output_ro.xml', './config_output_ro_removed.xml'
        )

        # Add this host as a read-write node (local)
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(self.LOCAL_NODE, single_node_xml, self.tmp_files[0])

        # Mock _rebalance_dbroots and _move_primary_node (only after the first node is added)
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ), patch(
            'cmapi_server.node_manipulation._rebalance_dbroots'
        ) as mock_rebalance_dbroots, patch(
            'cmapi_server.node_manipulation._move_primary_node'
        ) as mock_move_primary_node, patch(
            'cmapi_server.node_manipulation.update_dbroots_of_read_replicas'
        ) as mock_update_dbroots_of_read_replicas:
            # Add a read replica
            node_manipulation.add_node(
                self.REMOTE_NODE, self.tmp_files[0], self.tmp_files[1], read_replica=True
            )

            nc = NodeConfig()
            root = nc.get_current_config_root(self.tmp_files[1])

            # Check if PMS was added
            pms_node_ipaddr = root.find('./PMS2/IPAddr')
            self.assertEqual(pms_node_ipaddr.text, self.REMOTE_IP)

            # Check that WriteEngineServer was not added
            wes_node = root.find('./pm2_WriteEngineServer')
            self.assertIsNone(wes_node)

            # Check that get_read_replicas_pm_nums infers the read replica correctly
            read_replicas_pm_nums = nc.get_read_replicas_pm_nums(root)
            self.assertEqual(len(read_replicas_pm_nums), 1)
            self.assertEqual(read_replicas_pm_nums[0], 2)

            mock_rebalance_dbroots.assert_not_called()
            mock_move_primary_node.assert_not_called()
            # With a read replica added, update should be called
            mock_update_dbroots_of_read_replicas.assert_called_once()
            mock_update_dbroots_of_read_replicas.reset_mock()

            # Test read replica removal
            with (
                make_local_resolution_builder()
                .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
                .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
                .build()
            ):

                node_manipulation.remove_node(self.REMOTE_NODE, self.tmp_files[1], self.tmp_files[2])

            nc = NodeConfig()
            root = nc.get_current_config_root(self.tmp_files[2])
            read_replicas_pm_nums = nc.get_read_replicas_pm_nums(root)
            self.assertEqual(len(read_replicas_pm_nums), 0)

            for section_path in ['InactiveNodes', 'DesiredNodes', 'ActiveNodes']:
                section = root.find(section_path)
                self.assertFalse(any(n.text == self.REMOTE_IP for n in section.findall('./Node')))

            mock_rebalance_dbroots.assert_not_called()
            mock_move_primary_node.assert_not_called()
            mock_update_dbroots_of_read_replicas.assert_not_called()

    def test_add_dbroots_nodes_rebalance(self):
        # Toggle shared storage as requested and remember to restore it back.
        original_shared_storage = self._set_shared_storage(True)
        self.tmp_files = (
            './extra-dbroots-0.xml', './extra-dbroots-1.xml', './extra-dbroots-2.xml'
        )
        # add 2 dbroots, let's see what happens
        nc = NodeConfig()
        root = nc.get_current_config_root(tmp_mcs_config_filename)

        sysconf_node = root.find('./SystemConfig')
        dbroot_count_node = sysconf_node.find('./DBRootCount')
        dbroot_count = int(dbroot_count_node.text) + 2
        dbroot_count_node.text = str(dbroot_count)
        etree.SubElement(sysconf_node, 'DBRoot2').text = '/dummy_path/data2'
        etree.SubElement(sysconf_node, 'DBRoot10').text = '/dummy_path/data10'
        nc.write_config(root, self.tmp_files[0])

        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(self.LOCAL_NODE, self.tmp_files[0], self.tmp_files[1])

        # Do eyeball verification for now.
        nc = NodeConfig()
        root = nc.get_current_config_root(self.tmp_files[1])
        node = root.find('./PMS2/IPAddr')
        self.assertEqual(node.text, self.LOCAL_IP)

        hostname = socket.gethostname()
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.remove_node(hostname, self.tmp_files[1], self.tmp_files[2], test_mode=True)

        _ = self._set_shared_storage(original_shared_storage)

    def test_add_dbroot(self):
        self.tmp_files = (
            './dbroot-test0.xml', './dbroot-test1.xml', './dbroot-test2.xml',
            './dbroot-test3.xml', './dbroot-test4.xml'
        )
        # add a dbroot, verify it exists
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            id = node_manipulation.add_dbroot(tmp_mcs_config_filename, self.tmp_files[0])
        self.assertEqual(id, 2)
        nc = NodeConfig()
        root = nc.get_current_config_root(self.tmp_files[0])
        self.assertEqual(2, int(root.find('./SystemConfig/DBRootCount').text))
        self.assertEqual(f'{MCS_DATA_PATH}/data2', root.find('./SystemConfig/DBRoot2').text)

        # add a node, verify we can add a dbroot to each of them
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(self.LOCAL_NODE, tmp_mcs_config_filename, self.tmp_files[1])
            node_manipulation.add_node(self.REMOTE_IP, self.tmp_files[1], self.tmp_files[2])

            id1 = node_manipulation.add_dbroot(self.tmp_files[2], self.tmp_files[3], host=self.LOCAL_NODE)
            id2 = node_manipulation.add_dbroot(self.tmp_files[3], self.tmp_files[4], host=self.REMOTE_IP)
        self.assertEqual(id1, 2)
        self.assertEqual(id2, 3)

        root = nc.get_current_config_root(self.tmp_files[4])
        dbroot_count1 = int(root.find('./SystemModuleConfig/ModuleDBRootCount1-3').text)
        dbroot_count2 = int(root.find('./SystemModuleConfig/ModuleDBRootCount2-3').text)
        self.assertEqual(dbroot_count1 + dbroot_count2, 3)

        unique_dbroots = set()
        for i in range(1, dbroot_count1 + 1):
            unique_dbroots.add(int(root.find(f'./SystemModuleConfig/ModuleDBRootID1-{i}-3').text))
        for i in range(1, dbroot_count2 + 1):
            unique_dbroots.add(int(root.find(f'./SystemModuleConfig/ModuleDBRootID2-{i}-3').text))

        self.assertEqual(list(unique_dbroots), [1, 2, 3])

    def test_change_primary_node(self):
        # add a node, make it the primary, verify expected result
        self.tmp_files = ('./primary-node0.xml', './primary-node1.xml')
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(self.LOCAL_NODE, tmp_mcs_config_filename, self.tmp_files[0])

            node_manipulation.move_primary_node(self.tmp_files[0], self.tmp_files[1])

        root = NodeConfig().get_current_config_root(self.tmp_files[1])

        self.assertEqual(root.find('./ExeMgr1/IPAddr').text, self.LOCAL_IP)
        self.assertEqual(root.find('./DMLProc/IPAddr').text, self.LOCAL_IP)
        self.assertEqual(root.find('./DDLProc/IPAddr').text, self.LOCAL_IP)
        # This version doesn't support IPv6
        dbrm_controller_ip = root.find('./DBRM_Controller/IPAddr').text
        self.assertEqual(dbrm_controller_ip, self.LOCAL_NODE)
        self.assertEqual(root.find('./PrimaryNode').text, self.LOCAL_NODE)

    def test_unassign_dbroot1(self):
        self.tmp_files = ('./tud-0.xml', './tud-1.xml', './tud-2.xml', './tud-3.xml')
        with (
            make_local_resolution_builder()
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.REMOTE_NODE, self.REMOTE_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(self.LOCAL_NODE, tmp_mcs_config_filename, self.tmp_files[0])

        root = NodeConfig().get_current_config_root(self.tmp_files[0])
        (name, addr) = node_manipulation.find_dbroot1(root)
        self.assertEqual(name, self.LOCAL_NODE)

        # add a second node and more dbroots to make the test slightly more robust
        with (
            make_local_resolution_builder()
            .add_mapping(socket.gethostname(), self.LOCAL_IP, bidirectional=True)
            .add_mapping(self.LOCAL_NODE, self.LOCAL_IP, bidirectional=True)
            .build()
        ):
            node_manipulation.add_node(socket.gethostname(), self.tmp_files[0], self.tmp_files[1])

            node_manipulation.add_dbroot(self.tmp_files[1], self.tmp_files[2], socket.gethostname())
            node_manipulation.add_dbroot(self.tmp_files[2], self.tmp_files[3], self.LOCAL_NODE)

        root = NodeConfig().get_current_config_root(self.tmp_files[3])
        (name, addr) = node_manipulation.find_dbroot1(root)
        self.assertEqual(name, self.LOCAL_NODE)

        node_manipulation.unassign_dbroot1(root)
        caught_it = False
        try:
            node_manipulation.find_dbroot1(root)
        except node_manipulation.NodeNotFoundException:
            caught_it = True

        self.assertTrue(caught_it)

    def test_add_node_hostname_reverse_mismatch(self):
        """Adding a node by hostname should fail if reverse DNS doesn't map
        back to the provided hostname (neither exact nor FQDN starting with it).
        """
        self.tmp_files = ('./rev-mismatch-0.xml',)
        bad_hostname = 'badhost'

        with MockResolutionBuilder().add_mapping(bad_hostname, '10.0.0.5', bidirectional=False).build():
            with self.assertRaises(CMAPIBasicError):
                node_manipulation.add_node(bad_hostname, tmp_mcs_config_filename, self.tmp_files[0])

    def test_rebalance_noop_when_shared_storage_off(self):
        # Ensure shared storage flag is False
        current_cfg = AppStatefulConfig.get_config_copy()
        if current_cfg.flags.shared_storage_on:
            new_cfg = StatefulConfigModel(
                version=current_cfg.version.next_seq(),
                flags=StatefulFlagsModel(shared_storage_on=False)
            )
            AppStatefulConfig.apply_update(new_cfg)

        nc = NodeConfig()
        root = nc.get_current_config_root(tmp_mcs_config_filename)
        smc_before = etree.tostring(root.find('./SystemModuleConfig'))

        node_manipulation._rebalance_dbroots(root)

        smc_after = etree.tostring(root.find('./SystemModuleConfig'))
        self.assertEqual(smc_before, smc_after)


class TestDBRootsManipulation(unittest.TestCase):
    our_module_idx = 3
    ro_node1_ip = '192.168.1.3'
    ro_node2_ip = '192.168.1.4'

    def setUp(self):
        # Mock initial XML structure (add two nodes and two dbroots)
        self.root = etree.Element('Columnstore')
        # Add two PM modules with IP addresses
        smc = etree.SubElement(self.root, 'SystemModuleConfig')
        module_count = etree.SubElement(smc, 'ModuleCount3')
        module_count.text = '2'
        module1_ip = etree.SubElement(smc, 'ModuleIPAddr1-1-3')
        module1_ip.text = '192.168.1.1'
        module2_ip = etree.SubElement(smc, 'ModuleIPAddr2-1-3')
        module2_ip.text = '192.168.1.2'

        system_config = etree.SubElement(self.root, 'SystemConfig')
        dbroot_count = etree.SubElement(system_config, 'DBRootCount')
        dbroot_count.text = '2'
        dbroot1 = etree.SubElement(system_config, 'DBRoot1')
        dbroot1.text = '/data/dbroot1'
        dbroot2 = etree.SubElement(system_config, 'DBRoot2')
        dbroot2.text = '/data/dbroot2'

    def test_get_pm_module_num_to_addr_map(self):
        result = node_manipulation.get_pm_module_num_to_addr_map(self.root)

        expected = {
            1: '192.168.1.1',
            2: '192.168.1.2',
        }
        self.assertEqual(result, expected)

    def test_add_dbroots_of_other_nodes(self):
        '''add_dbroots_of_other_nodes must add dbroots of other nodes into mapping of the node.'''
        add_dbroots_of_other_nodes(self.root, self.our_module_idx)

        # Check that ModuleDBRootCount of the module was updated
        module_count = self.root.find(f'./SystemModuleConfig/ModuleDBRootCount{self.our_module_idx}-3')
        self.assertIsNotNone(module_count)
        self.assertEqual(module_count.text, '2')

        # Check that dbroots were added to ModuleDBRootID{module_num}-{i}-3
        dbroot1 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{self.our_module_idx}-1-3')
        dbroot2 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{self.our_module_idx}-2-3')
        self.assertIsNotNone(dbroot1)
        self.assertIsNotNone(dbroot2)
        self.assertEqual(dbroot1.text, '1')
        self.assertEqual(dbroot2.text, '2')

    def test_remove_dbroots_of_node(self):
        '''Test that remove_dbroots_of_node correctly removes dbroots from the node's mapping'''
        # Add dbroot association to the node
        smc = self.root.find('./SystemModuleConfig')
        dbroot1 = etree.SubElement(smc, f'ModuleDBRootID{self.our_module_idx}-1-3')
        dbroot1.text = '1'
        dbroot2 = etree.SubElement(smc, f'ModuleDBRootID{self.our_module_idx}-2-3')
        dbroot2.text = '2'
        # Add ModuleDBRootCount to the node
        module_count = etree.SubElement(smc, f'ModuleDBRootCount{self.our_module_idx}-3')
        module_count.text = '2'

        remove_dbroots_of_node(self.root, self.our_module_idx)

        # Check that ModuleDBRootCount was removed
        module_count = self.root.find(f'./SystemModuleConfig/ModuleDBRootCount{self.our_module_idx}-3')
        self.assertIsNone(module_count)
        # Check that dbroot mappings of the module were removed
        dbroot1 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{self.our_module_idx}-1-3')
        dbroot2 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{self.our_module_idx}-2-3')
        self.assertIsNone(dbroot1)
        self.assertIsNone(dbroot2)

    def test_update_dbroots_of_replica_nodes(self):
        """Test that update_dbroots_of_replica_nodes adds all existing dbroots to
        all existing replica nodes"""
        # Add two new new modules to the XML structure (two already exist)
        smc = self.root.find('./SystemModuleConfig')
        module_count = smc.find('./ModuleCount3')
        module_count.text = '4'
        module3_ip = etree.SubElement(smc, 'ModuleIPAddr3-1-3')
        module3_ip.text = self.ro_node1_ip
        module4_ip = etree.SubElement(smc, 'ModuleIPAddr4-1-3')
        module4_ip.text = self.ro_node2_ip

        # Mark modules 1 and 2 as write-enabled by adding WriteEngineServer sections
        wes1 = etree.SubElement(self.root, 'pm1_WriteEngineServer')
        etree.SubElement(wes1, 'IPAddr').text = '192.168.1.1'
        etree.SubElement(wes1, 'Port').text = '8630'
        wes2 = etree.SubElement(self.root, 'pm2_WriteEngineServer')
        etree.SubElement(wes2, 'IPAddr').text = '192.168.1.2'
        etree.SubElement(wes2, 'Port').text = '8630'

        update_dbroots_of_read_replicas(self.root)

        # Check that read replicas have all the dbroots
        for ro_module_idx in range(3, 5):
            module_count = self.root.find(f'./SystemModuleConfig/ModuleDBRootCount{ro_module_idx}-3')
            self.assertIsNotNone(module_count)
            self.assertEqual(module_count.text, '2')

            dbroot1 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{ro_module_idx}-1-3')
            dbroot2 = self.root.find(f'./SystemModuleConfig/ModuleDBRootID{ro_module_idx}-2-3')
            self.assertIsNotNone(dbroot1)
            self.assertIsNotNone(dbroot2)
            self.assertEqual(dbroot1.text, '1')
            self.assertEqual(dbroot2.text, '2')



import datetime
import logging
import sys
import unittest
from cmapi_server.logging_management import add_logging_level


class DatedTextTestResult(unittest.TextTestResult):
    def startTest(self, test: unittest.case.TestCase):
        self.stream.write('\n')
        self.stream.write(
            datetime.datetime.now().strftime("[%Y-%m-%d %H:%M:%S]: ")
        )
        return super().startTest(test)


def run_tests_from_package(p_name: str):
    logging.info(f'Starting tests from package {p_name}')
    loader = unittest.TestLoader()
    testsuite = loader.discover(
        pattern='test_*.py', start_dir=p_name, top_level_dir='./'
    )
    runner = unittest.runner.TextTestRunner(
        verbosity=3, failfast=True, resultclass=DatedTextTestResult
    )
    result = runner.run(testsuite)
    failed = False
    if not result.wasSuccessful():
        failed = True
        sys.exit(failed)
    logging.info(f'Finished tests from package {p_name}')


if __name__ == "__main__":
    add_logging_level('TRACE', 5)
    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s [%(levelname)s] (%(name)s) %(message)s'
    )
    run_tests_from_package('cmapi_server')
    run_tests_from_package('mcs_node_control')



# this class handles the comm with the agent; whatever it will be

import datetime
import logging
import threading
import time


logger = logging.getLogger('agent_comm')


# First an agent base class
class AgentBase:

    def activateNodes(self, nodes):
        print("AgentBase: Got activateNodes({})".format(nodes))

    def deactivateNodes(self, nodes):
        print("AgentBase: Got deactivateNodes({})".format(nodes))

    def movePrimaryNode(self, placeholder):
        print("AgentBase: Got movePrimaryNode()")

    def enterStandbyMode(self):
        print("AgentBase: Got enterStandbyMode()")

    def getNodeHealth(self):
        print("AgentBase: Got getNodeHealth()")
        return 0

    def raiseAlarm(self, msg):
        print("AgentBase: Got raiseAlarm({})".format(msg))

    def startTransaction(self, extra_nodes = [], remove_nodes = []):
        print(f"AgentBase: Got startTransaction, extra_nodes={extra_nodes}, remove_nodes={remove_nodes}")
        return 0

    def commitTransaction(self, txnid, nodes):
        print("AgentBase: Got commitTransaction")

    def rollbackTransaction(self, txnid, nodes):
        print("AgentBase: Got abortTransaction")



class OpAndArgs:
    name = None   # a callable in AgentBase
    args = None   # a tuple containing the args for the callable

    def __init__(self, name, *args):
        self.name = name
        self.args = args

    def __str__(self):
        return f"{str(self.name.__qualname__)}{str(self.args)}"

    def __hash__(self):
        return hash((self.name.__qualname__, str(self.args)))

    def __eq__(self, other):
        return self.name == other.name and self.args == other.args

    def __ne__(self, other):
        return not self.__eq__(other)

    def run(self):
        self.name(*self.args)


# The AgentComm class
# Doesn't do anything but pass along events to the Agent yet
# TODO: implement an event queue and a thread to pluck events and issue them
# to the agent. Done?
# TODO: de-dup events as they come in from the node monitor,
#       add to the event queue\
# TODO: rewrite using builtin Queue class
class AgentComm:

    def __init__(self, agent = None):
        if agent is None:
            self._agent = AgentBase()
        else:
            self._agent = agent

        # deduper contains queue contents, events in progress, and finished
        # events up to 10s after they finished
        self._deduper = {}
        self._die = False
        self._queue = []
        self._mutex = threading.Lock()
        self._thread = None

    def __del__(self):
        self.die()

    def start(self):
        self._die = False
        self._thread = threading.Thread(target=self._runner, name='AgentComm')
        self._thread.start()

    # TODO: rename to stop
    def die(self):
        self._die = True
        self._thread.join()

    # returns (len-of-event-queue, len-of-deduper)
    def getQueueSize(self):
        self._mutex.acquire()
        ret = (len(self._queue), len(self._deduper))
        self._mutex.release()
        return ret

    def activateNodes(self, nodes):
        self._addEvent(self._agent.activateNodes, (nodes))

    def deactivateNodes(self, nodes):
        self._addEvent(self._agent.deactivateNodes, (nodes))

    def movePrimaryNode(self):
        self._addEvent(self._agent.movePrimaryNode, ())

    def enterStandbyMode(self):
        # The other events are moot if this node has to enter standby mode
        self._mutex.acquire()
        op = OpAndArgs(self._agent.enterStandbyMode, ())
        self._queue = [ op ]
        self._deduper = { op : datetime.datetime.now() }
        self._mutex.release()

    def getNodeHealth(self):
        return self._agent.getNodeHealth()

    def raiseAlarm(self, msg):
        self._agent.raiseAlarm(msg)

    def _addEvent(self, name, args):
        """Interface to the event queue."""
        op = OpAndArgs(name, args)

        self._mutex.acquire()
        if op not in self._deduper:
            self._deduper[op] = None
            self._queue.append(op)
        self._mutex.release()

    def _getEvents(self):
        """
            This gets all queued events at once and prunes events older than
            10 seconds from the deduper.
        """
        self._mutex.acquire()
        ret = self._queue
        self._queue = []

        # prune events that finished more than 10 secs ago from the deduper
        tenSecsAgo = datetime.datetime.now() - datetime.timedelta(seconds = 10)
        for (op, finishTime) in list(self._deduper.items()):
            if finishTime is not None and finishTime < tenSecsAgo:
                del self._deduper[op]

        self._mutex.release()
        return ret

    def _requeueEvents(self, events):
        self._mutex.acquire()
        # events has commands issued before what is currently in _queue
        events.extend(self._queue)
        self._queue = events
        self._mutex.release()

    def _markEventsFinished(self, events):
        self._mutex.acquire()
        now = datetime.datetime.now()
        for event in events:
            self._deduper[event] = now
        self._mutex.release()

    def _runner(self):
        while not self._die:
            try:
                self.__runner()
            except Exception:
                logger.error(
                    'AgentComm.runner(): got an unrecognised exception.',
                    exc_info=True
                )
            if not self._die:
                time.sleep(1)
        logger.info('AgentComm.runner() exiting normally...')

    def __runner(self):
        while not self._die:
            events = self._getEvents()
            logger.trace(f'Get events from queue "{events}".')
            if len(events) == 0:
                time.sleep(5)
                continue

            nextPollTime = datetime.datetime.now() + datetime.timedelta(seconds = 5)

            nodes_added = set()
            nodes_removed = set()

            # scan the list of events, put together the extra_nodes and remove_nodes parameters to
            # startTransaction().  Note, we could consolidate the activate / deactivate calls here,
            # but that's a minor optimization not worth doing yet.
            needs_transaction = False
            for event in events:  # TODO: combine with loop below.
                #print(f"got event: {event}")

                # determine whether we need a transaction at all.
                # List the fcns that require a txn here.
                if not needs_transaction and event.name in (
                    self._agent.activateNodes,
                    self._agent.deactivateNodes,
                    self._agent.movePrimaryNode
                ):
                    needs_transaction = True

                if event.name == self._agent.activateNodes:
                    nodes = event.args[0]
                    for node in nodes:
                        nodes_added.add(node)
                elif event.name == self._agent.deactivateNodes:
                    nodes = event.args[0]
                    for node in nodes:
                        nodes_removed.add(node)

            if needs_transaction:
                logger.debug(
                    'Failover starts transaction to run upcoming event.'
                )
                (txn_id, nodes) = self._agent.startTransaction(
                    extra_nodes=list(nodes_added),
                    remove_nodes=list(nodes_removed)
                )

            # The problem with this is that it's all-or-nothing
            # It would be preferable to commit what has been done up to the point of failure
            # and discard the event that failed.
            # If the problem is with the event itself, then it may keep happening and block all
            # progress.
            try:
                for event in events:
                    #print(f"Running {event}")
                    event.run()
            except Exception as e:
                logger.error(
                    'AgentComm.runner(): got an unrecognised exception.',
                    exc_info=True
                )
                if needs_transaction:
                    logger.warning(
                        f'Aborting transaction {txn_id}',
                        exc_info=True
                    )
                    self._agent.rollbackTransaction(txn_id, nodes=nodes)
                # on failure, requeue the events in this batch to pick them up
                # again on the next iteration
                self._requeueEvents(events)
            else:
                if needs_transaction:
                    self._agent.commitTransaction(txn_id, nodes = nodes)
                self._markEventsFinished(events)
            finishTime = datetime.datetime.now()
            if nextPollTime > finishTime:
                time.sleep((nextPollTime - finishTime).seconds)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import sys
import time
import random
import re
import tempfile
import subprocess
import shutil
import argparse

# params overwrite priority:
#   for default:
#       default_params < {blackbox,whitebox}_default_params < args
#   for simple:
#       default_params < {blackbox,whitebox}_default_params <
#       simple_default_params <
#       {blackbox,whitebox}_simple_default_params < args
#   for cf_consistency:
#       default_params < {blackbox,whitebox}_default_params <
#       cf_consistency_params < args
#   for txn:
#       default_params < {blackbox,whitebox}_default_params < txn_params < args


default_params = {
    "acquire_snapshot_one_in": 10000,
    "backup_max_size": 100 * 1024 * 1024,
    # Consider larger number when backups considered more stable
    "backup_one_in": 100000,
    "batch_protection_bytes_per_key": lambda: random.choice([0, 8]),
    "block_size": 16384,
    "bloom_bits": lambda: random.choice([random.randint(0,19),
                                         random.lognormvariate(2.3, 1.3)]),
    "cache_index_and_filter_blocks": lambda: random.randint(0, 1),
    "cache_size": 1048576,
    "checkpoint_one_in": 1000000,
    "compression_type": lambda: random.choice(
        ["none", "snappy", "zlib", "bzip2", "lz4", "lz4hc", "xpress", "zstd"]),
    "bottommost_compression_type": lambda:
        "disable" if random.randint(0, 1) == 0 else
        random.choice(
            ["none", "snappy", "zlib", "bzip2", "lz4", "lz4hc", "xpress",
             "zstd"]),
    "checksum_type" : lambda: random.choice(["kCRC32c", "kxxHash", "kxxHash64", "kXXH3"]),
    "compression_max_dict_bytes": lambda: 16384 * random.randint(0, 1),
    "compression_zstd_max_train_bytes": lambda: 65536 * random.randint(0, 1),
    # Disabled compression_parallel_threads as the feature is not stable
    # lambda: random.choice([1] * 9 + [4])
    "compression_parallel_threads": 1,
    "compression_max_dict_buffer_bytes": lambda: (1 << random.randint(0, 40)) - 1,
    "clear_column_family_one_in": 0,
    "compact_files_one_in": 1000000,
    "compact_range_one_in": 1000000,
    "delpercent": 4,
    "delrangepercent": 1,
    "destroy_db_initially": 0,
    "enable_pipelined_write": lambda: random.randint(0, 1),
    "enable_compaction_filter": lambda: random.choice([0, 0, 0, 1]),
    "expected_values_dir": lambda: setup_expected_values_dir(),
    "fail_if_options_file_error": lambda: random.randint(0, 1),
    "flush_one_in": 1000000,
    "file_checksum_impl": lambda: random.choice(["none", "crc32c", "xxh64", "big"]),
    "get_live_files_one_in": 1000000,
    # Note: the following two are intentionally disabled as the corresponding
    # APIs are not guaranteed to succeed.
    "get_sorted_wal_files_one_in": 0,
    "get_current_wal_file_one_in": 0,
    # Temporarily disable hash index
    "index_type": lambda: random.choice([0, 0, 0, 2, 2, 3]),
    "iterpercent": 10,
    "mark_for_compaction_one_file_in": lambda: 10 * random.randint(0, 1),
    "max_background_compactions": 20,
    "max_bytes_for_level_base": 10485760,
    "max_key": 100000000,
    "max_write_buffer_number": 3,
    "mmap_read": lambda: random.randint(0, 1),
    # Setting `nooverwritepercent > 0` is only possible because we do not vary
    # the random seed, so the same keys are chosen by every run for disallowing
    # overwrites.
    "nooverwritepercent": 1,
    "open_files": lambda : random.choice([-1, -1, 100, 500000]),
    "optimize_filters_for_memory": lambda: random.randint(0, 1),
    "partition_filters": lambda: random.randint(0, 1),
    "partition_pinning": lambda: random.randint(0, 3),
    "pause_background_one_in": 1000000,
    "prefix_size" : lambda: random.choice([-1, 1, 5, 7, 8]),
    "prefixpercent": 5,
    "progress_reports": 0,
    "readpercent": 45,
    "recycle_log_file_num": lambda: random.randint(0, 1),
    "snapshot_hold_ops": 100000,
    "sst_file_manager_bytes_per_sec": lambda: random.choice([0, 104857600]),
    "sst_file_manager_bytes_per_truncate": lambda: random.choice([0, 1048576]),
    "long_running_snapshots": lambda: random.randint(0, 1),
    "subcompactions": lambda: random.randint(1, 4),
    "target_file_size_base": 2097152,
    "target_file_size_multiplier": 2,
    "test_batches_snapshots": lambda: random.randint(0, 1),
    "top_level_index_pinning": lambda: random.randint(0, 3),
    "unpartitioned_pinning": lambda: random.randint(0, 3),
    "use_direct_reads": lambda: random.randint(0, 1),
    "use_direct_io_for_flush_and_compaction": lambda: random.randint(0, 1),
    "mock_direct_io": False,
    "use_clock_cache": 0, # currently broken
    "use_full_merge_v1": lambda: random.randint(0, 1),
    "use_merge": lambda: random.randint(0, 1),
    # 999 -> use Bloom API
    "ribbon_starting_level": lambda: random.choice([random.randint(-1, 10), 999]),
    "use_block_based_filter": lambda: random.randint(0, 1),
    "verify_checksum": 1,
    "write_buffer_size": 4 * 1024 * 1024,
    "writepercent": 35,
    "format_version": lambda: random.choice([2, 3, 4, 5, 5]),
    "index_block_restart_interval": lambda: random.choice(range(1, 16)),
    "use_multiget" : lambda: random.randint(0, 1),
    "periodic_compaction_seconds" :
        lambda: random.choice([0, 0, 1, 2, 10, 100, 1000]),
    "compaction_ttl" : lambda: random.choice([0, 0, 1, 2, 10, 100, 1000]),
    # Test small max_manifest_file_size in a smaller chance, as most of the
    # time we wnat manifest history to be preserved to help debug
    "max_manifest_file_size" : lambda : random.choice(
        [t * 16384 if t < 3 else 1024 * 1024 * 1024 for t in range(1, 30)]),
    # Sync mode might make test runs slower so running it in a smaller chance
    "sync" : lambda : random.choice(
        [1 if t == 0 else 0 for t in range(0, 20)]),
    # Disable compation_readahead_size because the test is not passing.
    #"compaction_readahead_size" : lambda : random.choice(
    #    [0, 0, 1024 * 1024]),
    "db_write_buffer_size" : lambda: random.choice(
        [0, 0, 0, 1024 * 1024, 8 * 1024 * 1024, 128 * 1024 * 1024]),
    "avoid_unnecessary_blocking_io" : random.randint(0, 1),
    "write_dbid_to_manifest" : random.randint(0, 1),
    "avoid_flush_during_recovery" : random.choice(
        [1 if t == 0 else 0 for t in range(0, 8)]),
    "max_write_batch_group_size_bytes" : lambda: random.choice(
        [16, 64, 1024 * 1024, 16 * 1024 * 1024]),
    "level_compaction_dynamic_level_bytes" : True,
    "verify_checksum_one_in": 1000000,
    "verify_db_one_in": 100000,
    "continuous_verification_interval" : 0,
    "max_key_len": 3,
    "key_len_percent_dist": "1,30,69",
    "read_fault_one_in": lambda: random.choice([0, 32, 1000]),
    "open_metadata_write_fault_one_in": lambda: random.choice([0, 0, 8]),
    "open_write_fault_one_in": lambda: random.choice([0, 0, 16]),
    "open_read_fault_one_in": lambda: random.choice([0, 0, 32]),
    "sync_fault_injection": False,
    "get_property_one_in": 1000000,
    "paranoid_file_checks": lambda: random.choice([0, 1, 1, 1]),
    "max_write_buffer_size_to_maintain": lambda: random.choice(
        [0, 1024 * 1024, 2 * 1024 * 1024, 4 * 1024 * 1024, 8 * 1024 * 1024]),
    "user_timestamp_size": 0,
    "secondary_cache_fault_one_in" : lambda: random.choice([0, 0, 32]),
    "prepopulate_block_cache" : lambda: random.choice([0, 1]),
    "memtable_prefix_bloom_size_ratio": lambda: random.choice([0.001, 0.01, 0.1, 0.5]),
    "memtable_whole_key_filtering": lambda: random.randint(0, 1),
}

_TEST_DIR_ENV_VAR = 'TEST_TMPDIR'
_DEBUG_LEVEL_ENV_VAR = 'DEBUG_LEVEL'

stress_cmd = "./db_stress"

def is_release_mode():
    return os.environ.get(_DEBUG_LEVEL_ENV_VAR) == "0"


def get_dbname(test_name):
    test_dir_name = "rocksdb_crashtest_" + test_name
    test_tmpdir = os.environ.get(_TEST_DIR_ENV_VAR)
    if test_tmpdir is None or test_tmpdir == "":
        dbname = tempfile.mkdtemp(prefix=test_dir_name)
    else:
        dbname = test_tmpdir + "/" + test_dir_name
        shutil.rmtree(dbname, True)
        os.mkdir(dbname)
    return dbname

expected_values_dir = None
def setup_expected_values_dir():
    global expected_values_dir
    if expected_values_dir is not None:
        return expected_values_dir
    expected_dir_prefix = "rocksdb_crashtest_expected_"
    test_tmpdir = os.environ.get(_TEST_DIR_ENV_VAR)
    if test_tmpdir is None or test_tmpdir == "":
        expected_values_dir = tempfile.mkdtemp(
            prefix=expected_dir_prefix)
    else:
        # if tmpdir is specified, store the expected_values_dir under that dir
        expected_values_dir = test_tmpdir + "/rocksdb_crashtest_expected"
        if os.path.exists(expected_values_dir):
            shutil.rmtree(expected_values_dir)
        os.mkdir(expected_values_dir)
    return expected_values_dir


def is_direct_io_supported(dbname):
    with tempfile.NamedTemporaryFile(dir=dbname) as f:
        try:
            os.open(f.name, os.O_DIRECT)
        except BaseException:
            return False
        return True


blackbox_default_params = {
    "disable_wal": lambda: random.choice([0, 0, 0, 1]),
    # total time for this script to test db_stress
    "duration": 6000,
    # time for one db_stress instance to run
    "interval": 120,
    # since we will be killing anyway, use large value for ops_per_thread
    "ops_per_thread": 100000000,
    "reopen": 0,
    "set_options_one_in": 10000,
}

whitebox_default_params = {
    # TODO: enable this once we figure out how to adjust kill odds for WAL-
    # disabled runs, and either (1) separate full `db_stress` runs out of
    # whitebox crash or (2) support verification at end of `db_stress` runs
    # that ran with WAL disabled.
    "disable_wal": 0,
    "duration": 10000,
    "log2_keys_per_lock": 10,
    "ops_per_thread": 200000,
    "random_kill_odd": 888887,
    "reopen": 20,
}

simple_default_params = {
    "allow_concurrent_memtable_write": lambda: random.randint(0, 1),
    "column_families": 1,
    # TODO: re-enable once below loop succeeds for a while (a few minutes should
    # suffice):
    # `while rm -rf /dev/shm/single_stress && ./db_stress --clear_column_family_one_in=0 --column_families=1 --db=/dev/shm/single_stress --experimental_mempurge_threshold=5.493146827397074 --flush_one_in=10000 --reopen=0 --write_buffer_size=262144 --value_size_mult=33 --max_write_buffer_number=3 -ops_per_thread=10000; do : ; done`
    "experimental_mempurge_threshold": 0,
    "max_background_compactions": 1,
    "max_bytes_for_level_base": 67108864,
    "memtablerep": "skip_list",
    "target_file_size_base": 16777216,
    "target_file_size_multiplier": 1,
    "test_batches_snapshots": 0,
    "write_buffer_size": 32 * 1024 * 1024,
    "level_compaction_dynamic_level_bytes": False,
    "paranoid_file_checks": lambda: random.choice([0, 1, 1, 1]),
}

blackbox_simple_default_params = {
    "open_files": -1,
    "set_options_one_in": 0,
}

whitebox_simple_default_params = {}

cf_consistency_params = {
    "disable_wal": lambda: random.randint(0, 1),
    "reopen": 0,
    "test_cf_consistency": 1,
    # use small value for write_buffer_size so that RocksDB triggers flush
    # more frequently
    "write_buffer_size": 1024 * 1024,
    "enable_pipelined_write": lambda: random.randint(0, 1),
    # Snapshots are used heavily in this test mode, while they are incompatible
    # with compaction filter.
    "enable_compaction_filter": 0,
}

txn_params = {
    "use_txn" : 1,
    # Avoid lambda to set it once for the entire test
    "txn_write_policy": random.randint(0, 2),
    "unordered_write": random.randint(0, 1),
    # TODO: there is such a thing as transactions with WAL disabled. We should
    # cover that case.
    "disable_wal": 0,
    # OpenReadOnly after checkpoint is not currnetly compatible with WritePrepared txns
    "checkpoint_one_in": 0,
    # pipeline write is not currnetly compatible with WritePrepared txns
    "enable_pipelined_write": 0,
}

best_efforts_recovery_params = {
    "best_efforts_recovery": True,
    "skip_verifydb": True,
    "verify_db_one_in": 0,
    "continuous_verification_interval": 0,
}

blob_params = {
    "allow_setting_blob_options_dynamically": 1,
    # Enable blob files and GC with a 75% chance initially; note that they might still be
    # enabled/disabled during the test via SetOptions
    "enable_blob_files": lambda: random.choice([0] + [1] * 3),
    "min_blob_size": lambda: random.choice([0, 8, 16]),
    "blob_file_size": lambda: random.choice([1048576, 16777216, 268435456, 1073741824]),
    "blob_compression_type": lambda: random.choice(["none", "snappy", "lz4", "zstd"]),
    "enable_blob_garbage_collection": lambda: random.choice([0] + [1] * 3),
    "blob_garbage_collection_age_cutoff": lambda: random.choice([0.0, 0.25, 0.5, 0.75, 1.0]),
    "blob_garbage_collection_force_threshold": lambda: random.choice([0.5, 0.75, 1.0]),
    "blob_compaction_readahead_size": lambda: random.choice([0, 1048576, 4194304]),
}

ts_params = {
    "test_cf_consistency": 0,
    "test_batches_snapshots": 0,
    "user_timestamp_size": 8,
    "use_merge": 0,
    "use_full_merge_v1": 0,
    "use_txn": 0,
    "read_only": 0,
    "secondary_catch_up_one_in": 0,
    "continuous_verification_interval": 0,
    "checkpoint_one_in": 0,
    "enable_blob_files": 0,
    "use_blob_db": 0,
    "enable_compaction_filter": 0,
    "ingest_external_file_one_in": 0,
    "use_block_based_filter": 0,
}

def finalize_and_sanitize(src_params):
    dest_params = dict([(k,  v() if callable(v) else v)
                        for (k, v) in src_params.items()])
    if dest_params.get("compression_max_dict_bytes") == 0:
        dest_params["compression_zstd_max_train_bytes"] = 0
        dest_params["compression_max_dict_buffer_bytes"] = 0
    if dest_params.get("compression_type") != "zstd":
        dest_params["compression_zstd_max_train_bytes"] = 0
    if dest_params.get("allow_concurrent_memtable_write", 1) == 1:
        dest_params["memtablerep"] = "skip_list"
    if dest_params["mmap_read"] == 1:
        dest_params["use_direct_io_for_flush_and_compaction"] = 0
        dest_params["use_direct_reads"] = 0
    if (dest_params["use_direct_io_for_flush_and_compaction"] == 1
            or dest_params["use_direct_reads"] == 1) and \
            not is_direct_io_supported(dest_params["db"]):
        if is_release_mode():
            print("{} does not support direct IO. Disabling use_direct_reads and "
                    "use_direct_io_for_flush_and_compaction.\n".format(
                        dest_params["db"]))
            dest_params["use_direct_reads"] = 0
            dest_params["use_direct_io_for_flush_and_compaction"] = 0
        else:
            dest_params["mock_direct_io"] = True

    # DeleteRange is not currnetly compatible with Txns and timestamp
    if (dest_params.get("test_batches_snapshots") == 1 or
        dest_params.get("use_txn") == 1 or
        dest_params.get("user_timestamp_size") > 0):
        dest_params["delpercent"] += dest_params["delrangepercent"]
        dest_params["delrangepercent"] = 0
    # Only under WritePrepared txns, unordered_write would provide the same guarnatees as vanilla rocksdb
    if dest_params.get("unordered_write", 0) == 1:
        dest_params["txn_write_policy"] = 1
        dest_params["allow_concurrent_memtable_write"] = 1
    if dest_params.get("disable_wal", 0) == 1:
        dest_params["atomic_flush"] = 1
        # The `DbStressCompactionFilter` can apply memtable updates to SST
        # files, which would be problematic without WAL since such updates are
        # expected to be lost in crash recoveries.
        dest_params["enable_compaction_filter"] = 0
        dest_params["sync"] = 0
        dest_params["write_fault_one_in"] = 0
    if dest_params.get("open_files", 1) != -1:
        # Compaction TTL and periodic compactions are only compatible
        # with open_files = -1
        dest_params["compaction_ttl"] = 0
        dest_params["periodic_compaction_seconds"] = 0
    if dest_params.get("compaction_style", 0) == 2:
        # Disable compaction TTL in FIFO compaction, because right
        # now assertion failures are triggered.
        dest_params["compaction_ttl"] = 0
        dest_params["periodic_compaction_seconds"] = 0
    if dest_params["partition_filters"] == 1:
        if dest_params["index_type"] != 2:
            dest_params["partition_filters"] = 0
        else:
            dest_params["use_block_based_filter"] = 0
    if dest_params["ribbon_starting_level"] < 999:
        dest_params["use_block_based_filter"] = 0
    if dest_params.get("atomic_flush", 0) == 1:
        # disable pipelined write when atomic flush is used.
        dest_params["enable_pipelined_write"] = 0
    if dest_params.get("sst_file_manager_bytes_per_sec", 0) == 0:
        dest_params["sst_file_manager_bytes_per_truncate"] = 0
    if dest_params.get("enable_compaction_filter", 0) == 1:
        # Compaction filter is incompatible with snapshots. Need to avoid taking
        # snapshots, as well as avoid operations that use snapshots for
        # verification.
        dest_params["acquire_snapshot_one_in"] = 0
        dest_params["compact_range_one_in"] = 0
        # Give the iterator ops away to reads.
        dest_params["readpercent"] += dest_params.get("iterpercent", 10)
        dest_params["iterpercent"] = 0
        dest_params["test_batches_snapshots"] = 0
    if dest_params.get("prefix_size") == -1:
        dest_params["readpercent"] += dest_params.get("prefixpercent", 20)
        dest_params["prefixpercent"] = 0
        dest_params["test_batches_snapshots"] = 0
    if dest_params.get("test_batches_snapshots") == 0:
        dest_params["batch_protection_bytes_per_key"] = 0
    if (dest_params.get("prefix_size") == -1 and
        dest_params.get("memtable_whole_key_filtering") == 0):
        dest_params["memtable_prefix_bloom_size_ratio"] = 0
    return dest_params

def gen_cmd_params(args):
    params = {}

    params.update(default_params)
    if args.test_type == 'blackbox':
        params.update(blackbox_default_params)
    if args.test_type == 'whitebox':
        params.update(whitebox_default_params)
    if args.simple:
        params.update(simple_default_params)
        if args.test_type == 'blackbox':
            params.update(blackbox_simple_default_params)
        if args.test_type == 'whitebox':
            params.update(whitebox_simple_default_params)
    if args.cf_consistency:
        params.update(cf_consistency_params)
    if args.txn:
        params.update(txn_params)
    if args.test_best_efforts_recovery:
        params.update(best_efforts_recovery_params)
    if args.enable_ts:
        params.update(ts_params)

    # Best-effort recovery and BlobDB are currently incompatible. Test BE recovery
    # if specified on the command line; otherwise, apply BlobDB related overrides
    # with a 10% chance.
    if (not args.test_best_efforts_recovery and
        not args.enable_ts and
        random.choice([0] * 9 + [1]) == 1):
        params.update(blob_params)

    for k, v in vars(args).items():
        if v is not None:
            params[k] = v
    return params


def gen_cmd(params, unknown_params):
    finalzied_params = finalize_and_sanitize(params)
    cmd = [stress_cmd] + [
        '--{0}={1}'.format(k, v)
        for k, v in [(k, finalzied_params[k]) for k in sorted(finalzied_params)]
        if k not in set(['test_type', 'simple', 'duration', 'interval',
                         'random_kill_odd', 'cf_consistency', 'txn',
                         'test_best_efforts_recovery', 'enable_ts', 'stress_cmd'])
        and v is not None] + unknown_params
    return cmd


# Inject inconsistency to db directory.
def inject_inconsistencies_to_db_dir(dir_path):
    files = os.listdir(dir_path)
    file_num_rgx = re.compile(r'(?P<number>[0-9]{6})')
    largest_fnum = 0
    for f in files:
        m = file_num_rgx.search(f)
        if m and not f.startswith('LOG'):
            largest_fnum = max(largest_fnum, int(m.group('number')))

    candidates = [
        f for f in files if re.search(r'[0-9]+\.sst', f)
    ]
    deleted = 0
    corrupted = 0
    for f in candidates:
        rnd = random.randint(0, 99)
        f_path = os.path.join(dir_path, f)
        if rnd < 10:
            os.unlink(f_path)
            deleted = deleted + 1
        elif 10 <= rnd and rnd < 30:
            with open(f_path, "a") as fd:
                fd.write('12345678')
            corrupted = corrupted + 1
    print('Removed %d table files' % deleted)
    print('Corrupted %d table files' % corrupted)

    # Add corrupted MANIFEST and SST
    for num in range(largest_fnum + 1, largest_fnum + 10):
        rnd = random.randint(0, 1)
        fname = ("MANIFEST-%06d" % num) if rnd == 0 else ("%06d.sst" % num)
        print('Write %s' % fname)
        with open(os.path.join(dir_path, fname), "w") as fd:
            fd.write("garbage")

def execute_cmd(cmd, timeout):
    child = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                             stdout=subprocess.PIPE)
    print("Running db_stress with pid=%d: %s\n\n"
          % (child.pid, ' '.join(cmd)))

    try:
        outs, errs = child.communicate(timeout=timeout)
        hit_timeout = False
        print("WARNING: db_stress ended before kill: exitcode=%d\n"
              % child.returncode)
    except subprocess.TimeoutExpired:
        hit_timeout = True
        child.kill()
        print("KILLED %d\n" % child.pid)
        outs, errs = child.communicate()

    return hit_timeout, child.returncode, outs.decode('utf-8'), errs.decode('utf-8')


# This script runs and kills db_stress multiple times. It checks consistency
# in case of unsafe crashes in RocksDB.
def blackbox_crash_main(args, unknown_args):
    cmd_params = gen_cmd_params(args)
    dbname = get_dbname('blackbox')
    exit_time = time.time() + cmd_params['duration']

    print("Running blackbox-crash-test with \n"
          + "interval_between_crash=" + str(cmd_params['interval']) + "\n"
          + "total-duration=" + str(cmd_params['duration']) + "\n")

    while time.time() < exit_time:
        cmd = gen_cmd(dict(
            list(cmd_params.items())
            + list({'db': dbname}.items())), unknown_args)

        hit_timeout, retcode, outs, errs = execute_cmd(cmd, cmd_params['interval'])

        if not hit_timeout:
            print('Exit Before Killing')
            print('stdout:')
            print(outs)
            print('stderr:')
            print(errs)
            sys.exit(2)

        for line in errs.split('\n'):
            if line != '' and  not line.startswith('WARNING'):
                print('stderr has error message:')
                print('***' + line + '***')

        time.sleep(1)  # time to stabilize before the next run

        if args.test_best_efforts_recovery:
            inject_inconsistencies_to_db_dir(dbname)

        time.sleep(1)  # time to stabilize before the next run

    # we need to clean up after ourselves -- only do this on test success
    shutil.rmtree(dbname, True)


# This python script runs db_stress multiple times. Some runs with
# kill_random_test that causes rocksdb to crash at various points in code.
def whitebox_crash_main(args, unknown_args):
    cmd_params = gen_cmd_params(args)
    dbname = get_dbname('whitebox')

    cur_time = time.time()
    exit_time = cur_time + cmd_params['duration']
    half_time = cur_time + cmd_params['duration'] // 2

    print("Running whitebox-crash-test with \n"
          + "total-duration=" + str(cmd_params['duration']) + "\n")

    total_check_mode = 4
    check_mode = 0
    kill_random_test = cmd_params['random_kill_odd']
    kill_mode = 0

    while time.time() < exit_time:
        if check_mode == 0:
            additional_opts = {
                # use large ops per thread since we will kill it anyway
                "ops_per_thread": 100 * cmd_params['ops_per_thread'],
            }
            # run with kill_random_test, with three modes.
            # Mode 0 covers all kill points. Mode 1 covers less kill points but
            # increases change of triggering them. Mode 2 covers even less
            # frequent kill points and further increases triggering change.
            if kill_mode == 0:
                additional_opts.update({
                    "kill_random_test": kill_random_test,
                })
            elif kill_mode == 1:
                if cmd_params.get('disable_wal', 0) == 1:
                    my_kill_odd = kill_random_test // 50 + 1
                else:
                    my_kill_odd = kill_random_test // 10 + 1
                additional_opts.update({
                    "kill_random_test": my_kill_odd,
                    "kill_exclude_prefixes": "WritableFileWriter::Append,"
                    + "WritableFileWriter::WriteBuffered",
                })
            elif kill_mode == 2:
                # TODO: May need to adjust random odds if kill_random_test
                # is too small.
                additional_opts.update({
                    "kill_random_test": (kill_random_test // 5000 + 1),
                    "kill_exclude_prefixes": "WritableFileWriter::Append,"
                    "WritableFileWriter::WriteBuffered,"
                    "PosixMmapFile::Allocate,WritableFileWriter::Flush",
                })
            # Run kill mode 0, 1 and 2 by turn.
            kill_mode = (kill_mode + 1) % 3
        elif check_mode == 1:
            # normal run with universal compaction mode
            additional_opts = {
                "kill_random_test": None,
                "ops_per_thread": cmd_params['ops_per_thread'],
                "compaction_style": 1,
            }
            # Single level universal has a lot of special logic. Ensure we cover
            # it sometimes.
            if random.randint(0, 1) == 1:
                additional_opts.update({
                    "num_levels": 1,
                })
        elif check_mode == 2:
            # normal run with FIFO compaction mode
            # ops_per_thread is divided by 5 because FIFO compaction
            # style is quite a bit slower on reads with lot of files
            additional_opts = {
                "kill_random_test": None,
                "ops_per_thread": cmd_params['ops_per_thread'] // 5,
                "compaction_style": 2,
            }
        else:
            # normal run
            additional_opts = {
                "kill_random_test": None,
                "ops_per_thread": cmd_params['ops_per_thread'],
            }

        cmd = gen_cmd(dict(list(cmd_params.items())
            + list(additional_opts.items())
            + list({'db': dbname}.items())), unknown_args)

        print("Running:" + ' '.join(cmd) + "\n")  # noqa: E999 T25377293 Grandfathered in

        # If the running time is 15 minutes over the run time, explicit kill and
        # exit even if white box kill didn't hit. This is to guarantee run time
        # limit, as if it runs as a job, running too long will create problems
        # for job scheduling or execution.
        # TODO detect a hanging condition. The job might run too long as RocksDB
        # hits a hanging bug.
        hit_timeout, retncode, stdoutdata, stderrdata = execute_cmd(
            cmd, exit_time - time.time() + 900)
        msg = ("check_mode={0}, kill option={1}, exitcode={2}\n".format(
               check_mode, additional_opts['kill_random_test'], retncode))

        print(msg)
        print(stdoutdata)
        print(stderrdata)

        if hit_timeout:
            print("Killing the run for running too long")
            break

        expected = False
        if additional_opts['kill_random_test'] is None and (retncode == 0):
            # we expect zero retncode if no kill option
            expected = True
        elif additional_opts['kill_random_test'] is not None and retncode <= 0:
            # When kill option is given, the test MIGHT kill itself.
            # If it does, negative retncode is expected. Otherwise 0.
            expected = True

        if not expected:
            print("TEST FAILED. See kill option and exit code above!!!\n")
            sys.exit(1)

        stderrdata = stderrdata.lower()
        errorcount = (stderrdata.count('error') -
                      stderrdata.count('got errors 0 times'))
        print("#times error occurred in output is " + str(errorcount) +
                "\n")

        if (errorcount > 0):
            print("TEST FAILED. Output has 'error'!!!\n")
            sys.exit(2)
        if (stderrdata.find('fail') >= 0):
            print("TEST FAILED. Output has 'fail'!!!\n")
            sys.exit(2)

        # First half of the duration, keep doing kill test. For the next half,
        # try different modes.
        if time.time() > half_time:
            # we need to clean up after ourselves -- only do this on test
            # success
            shutil.rmtree(dbname, True)
            os.mkdir(dbname)
            cmd_params.pop('expected_values_dir', None)
            check_mode = (check_mode + 1) % total_check_mode

        time.sleep(1)  # time to stabilize after a kill


def main():
    global stress_cmd

    parser = argparse.ArgumentParser(description="This script runs and kills \
        db_stress multiple times")
    parser.add_argument("test_type", choices=["blackbox", "whitebox"])
    parser.add_argument("--simple", action="store_true")
    parser.add_argument("--cf_consistency", action='store_true')
    parser.add_argument("--txn", action='store_true')
    parser.add_argument("--test_best_efforts_recovery", action='store_true')
    parser.add_argument("--enable_ts", action='store_true')
    parser.add_argument("--stress_cmd")

    all_params = dict(list(default_params.items())
                      + list(blackbox_default_params.items())
                      + list(whitebox_default_params.items())
                      + list(simple_default_params.items())
                      + list(blackbox_simple_default_params.items())
                      + list(whitebox_simple_default_params.items())
                      + list(blob_params.items())
                      + list(ts_params.items()))

    for k, v in all_params.items():
        parser.add_argument("--" + k, type=type(v() if callable(v) else v))
    # unknown_args are passed directly to db_stress
    args, unknown_args = parser.parse_known_args()

    test_tmpdir = os.environ.get(_TEST_DIR_ENV_VAR)
    if test_tmpdir is not None and not os.path.isdir(test_tmpdir):
        print('%s env var is set to a non-existent directory: %s' %
                (_TEST_DIR_ENV_VAR, test_tmpdir))
        sys.exit(1)

    if args.stress_cmd:
        stress_cmd = args.stress_cmd
    if args.test_type == 'blackbox':
        blackbox_crash_main(args, unknown_args)
    if args.test_type == 'whitebox':
        whitebox_crash_main(args, unknown_args)
    # Only delete the `expected_values_dir` if test passes
    if expected_values_dir is not None:
        shutil.rmtree(expected_values_dir)


if __name__ == '__main__':
    main()



import json
import logging

import cherrypy

from cmapi_server.constants import _version
from cmapi_server.controllers.endpoints import (
    ApiKeyController, AppController, BeginController, ClusterController,
    CommitController, ConfigController, ExtentMapController,
    LoggingConfigController, NodeController, NodeProcessController,
    RollbackController, ShutdownController, StartController, StatusController,
    CmapiConfigController,
)
from cmapi_server.controllers.s3dataload import S3DataLoadController


dispatcher = cherrypy.dispatch.RoutesDispatcher()
logger = logging.getLogger(__name__)


# /_version/status (GET)
dispatcher.connect(name = 'status',
                   route = f'/cmapi/{_version}/node/status',
                   action = 'get_status',
                   controller = StatusController(),
                   conditions = {'method': ['GET']})


# /_version/master (GET)
dispatcher.connect(name = 'get_primary',
                   route = f'/cmapi/{_version}/node/primary',
                   action = 'get_primary',
                   controller = StatusController(),
                   conditions = {'method': ['GET']})


# /_version/new_primary (GET)
dispatcher.connect(name = 'get_new_primary',
                   route = f'/cmapi/{_version}/node/new_primary',
                   action = 'get_new_primary',
                   controller = StatusController(),
                   conditions = {'method': ['GET']})


# /_version/config/ (GET)
dispatcher.connect(name = 'get_config', # what does this name is used for?
                   route = f'/cmapi/{_version}/node/config',
                   action = 'get_config',
                   controller = ConfigController(),
                   conditions = {'method': ['GET']})


# /_version/config/ (PUT)
dispatcher.connect(name = 'put_config',
                   route = f'/cmapi/{_version}/node/config',
                   action = 'put_config',
                   controller = ConfigController(),
                   conditions = {'method': ['PUT']})


# /_version/begin/ (PUT)
dispatcher.connect(name = 'put_begin',
                   route = f'/cmapi/{_version}/node/begin',
                   action = 'put_begin',
                   controller = BeginController(),
                   conditions = {'method': ['PUT']})


# /_version/rollback/ (PUT)
dispatcher.connect(name = 'put_rollback',
                   route = f'/cmapi/{_version}/node/rollback',
                   action = 'put_rollback',
                   controller = RollbackController(),
                   conditions = {'method': ['PUT']})


# /_version/commit/ (PUT)
dispatcher.connect(name = 'put_commit',
                   route = f'/cmapi/{_version}/node/commit',
                   action = 'put_commit',
                   controller = CommitController(),
                   conditions = {'method': ['PUT']})


# /_version/start/ (PUT)
dispatcher.connect(name = 'start',
                   route = f'/cmapi/{_version}/node/start',
                   action = 'put_start',
                   controller = StartController(),
                   conditions = {'method': ['PUT']})


# /_version/shutdown/ (PUT)
dispatcher.connect(name = 'shutdown',
                   route = f'/cmapi/{_version}/node/shutdown',
                   action = 'put_shutdown',
                   controller = ShutdownController(),
                   conditions = {'method': ['PUT']})


# /_version/meta/em/ (GET)
dispatcher.connect(name = 'get_em',
                   route = f'/cmapi/{_version}/node/meta/em',
                   action = 'get_em',
                   controller = ExtentMapController(),
                   conditions = {'method': ['GET']})


# /_version/meta/journal/ (GET)
dispatcher.connect(name = 'get_journal',
                   route = f'/cmapi/{_version}/node/meta/journal',
                   action = 'get_journal',
                   controller = ExtentMapController(),
                   conditions = {'method': ['GET']})


# /_version/meta/vss/ (GET)
dispatcher.connect(name = 'get_vss',
                   route = f'/cmapi/{_version}/node/meta/vss',
                   action = 'get_vss',
                   controller = ExtentMapController(),
                   conditions = {'method': ['GET']})


# /_version/meta/vbbm/ (GET)
dispatcher.connect(name = 'get_vbbm',
                   route = f'/cmapi/{_version}/node/meta/vbbm',
                   action = 'get_vbbm',
                   controller = ExtentMapController(),
                   conditions = {'method': ['GET']})


# /_version/meta/footprint/ (GET)
dispatcher.connect(name = 'get_footprint',
                   route = f'/cmapi/{_version}/node/meta/footprint',
                   action = 'get_footprint',
                   controller = ExtentMapController(),
                   conditions = {'method': ['GET']})


# /_version/cluster/start/ (PUT)
dispatcher.connect(name = 'cluster_start',
                   route = f'/cmapi/{_version}/cluster/start',
                   action = 'put_start',
                   controller = ClusterController(),
                   conditions = {'method': ['PUT']})


# /_version/cluster/shutdown/ (PUT)
dispatcher.connect(name = 'cluster_shutdown',
                   route = f'/cmapi/{_version}/cluster/shutdown',
                   action = 'put_shutdown',
                   controller = ClusterController(),
                   conditions = {'method': ['PUT']})


# /_version/cluster/mode-set/ (PUT)
dispatcher.connect(name = 'cluster_mode_set',
                   route = f'/cmapi/{_version}/cluster/mode-set',
                   action = 'put_mode_set',
                   controller = ClusterController(),
                   conditions = {'method': ['PUT']})


# /_version/cluster/node/ (POST, PUT)
dispatcher.connect(name = 'cluster_add_node',
                   route = f'/cmapi/{_version}/cluster/node',
                   action = 'put_add_node',
                   controller = ClusterController(),
                   conditions = {'method': ['POST', 'PUT']})


# /_version/cluster/node/ (DELETE)
dispatcher.connect(name = 'cluster_remove_node',
                   route = f'/cmapi/{_version}/cluster/node',
                   action = 'delete_remove_node',
                   controller = ClusterController(),
                   conditions = {'method': ['DELETE']})


# /_version/cluster/status/ (GET)
dispatcher.connect(name = 'cluster_status',
                   route = f'/cmapi/{_version}/cluster/status',
                   action = 'get_status',
                   controller = ClusterController(),
                   conditions = {'method': ['GET']})


# /_version/node/apikey-set/ (PUT)
dispatcher.connect(
    name = 'node_set_api_key',
    route = f'/cmapi/{_version}/node/apikey-set',
    action = 'set_api_key',
    controller = ApiKeyController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/apikey-set/ (PUT)
dispatcher.connect(
    name = 'cluster_set_api_key',
    route = f'/cmapi/{_version}/cluster/apikey-set',
    action = 'set_api_key',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/load_s3data (POST, PUT)
dispatcher.connect(name = 'cluster_load_s3data',
                   route = f'/cmapi/{_version}/cluster/load_s3data',
                   action = 'load_s3data',
                   controller = S3DataLoadController(),
                   conditions = {'method': ['POST', 'PUT']})


# /_version/node/log-config/ (PUT)
dispatcher.connect(
    name = 'node_set_log_level',
    route = f'/cmapi/{_version}/node/log-level',
    action = 'set_log_level',
    controller = LoggingConfigController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/log-config'/ (PUT)
dispatcher.connect(
    name = 'cluster_set_log_level',
    route = f'/cmapi/{_version}/cluster/log-level',
    action = 'set_log_level',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /ready (GET)
dispatcher.connect(
    name = 'app_ready',
    route = '/cmapi/ready',
    action = 'ready',
    controller = AppController(),
    conditions = {'method': ['GET']}
)


# /_version/node/stop_dmlproc/ (PUT)
dispatcher.connect(
    name = 'stop_dmlproc',
    route = f'/cmapi/{_version}/node/stop_dmlproc',
    action = 'put_stop_dmlproc',
    controller = NodeProcessController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/is_process_running/ (GET)
dispatcher.connect(
    name = 'is_process_running',
    route = f'/cmapi/{_version}/node/is_process_running',
    action = 'get_process_running',
    controller = NodeProcessController(),
    conditions = {'method': ['GET']}
)


# /_version/cluster/health/ (PUT)
dispatcher.connect(
    name = 'cluster_get_health',
    route = f'/cmapi/{_version}/cluster/health',
    action = 'get_health',
    controller = ClusterController(),
    conditions = {'method': ['GET']}
)


# /_version/node/versions (GET)
dispatcher.connect(
    name = 'node_get_versions',
    route = f'/cmapi/{_version}/node/versions',
    action = 'get_versions',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/cluster/version (GET)
dispatcher.connect(
    name = 'cluster_get_versions',
    route = f'/cmapi/{_version}/cluster/versions',
    action = 'get_versions',
    controller = ClusterController(),
    conditions = {'method': ['GET']}
)


# /_version/node/latest-mdb-version (GET)
dispatcher.connect(
    name = 'get_latest_mdb_version',
    route = f'/cmapi/{_version}/node/latest-mdb-version',
    action = 'latest_mdb_version',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/node/validate-mdb-version (GET)
dispatcher.connect(
    name = 'get_validate_mdb_version',
    route = f'/cmapi/{_version}/node/validate-mdb-version',
    action = 'validate_mdb_version',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/node/validate-es-token (GET)
dispatcher.connect(
    name = 'get_validate_es_token',
    route = f'/cmapi/{_version}/node/validate-es-token',
    action = 'validate_es_token',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/node/stop-mariadb (PUT)
dispatcher.connect(
    name = 'node_stop_mariadb',
    route = f'/cmapi/{_version}/node/stop-mariadb',
    action = 'stop_mariadb',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/stop-mariadb (PUT)
dispatcher.connect(
    name = 'cluster_stop_mariadb',
    route = f'/cmapi/{_version}/cluster/stop-mariadb',
    action = 'stop_mariadb',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/start-mariadb (PUT)
dispatcher.connect(
    name = 'node_start_mariadb',
    route = f'/cmapi/{_version}/node/start-mariadb',
    action = 'start_mariadb',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/start-mariadb (PUT)
dispatcher.connect(
    name = 'cluster_start_mariadb',
    route = f'/cmapi/{_version}/cluster/start-mariadb',
    action = 'start_mariadb',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/install-repo (PUT)
dispatcher.connect(
    name = 'node_install_repo',
    route = f'/cmapi/{_version}/node/install-repo',
    action = 'install_repo',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/install-repo (PUT)
dispatcher.connect(
    name = 'cluster_install_repo',
    route = f'/cmapi/{_version}/cluster/install-repo',
    action = 'install_repo',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/repo-pkg-versions (GET)
dispatcher.connect(
    name = 'get_repo_pkg_versions',
    route = f'/cmapi/{_version}/node/repo-pkg-versions',
    action = 'repo_pkg_versions',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/node/preupgrade-backup (PUT)
dispatcher.connect(
    name = 'node_preupgrade_backup',
    route = f'/cmapi/{_version}/node/preupgrade-backup',
    action = 'preupgrade_backup',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/preupgrade-backup (PUT)
dispatcher.connect(
    name = 'cluster_preupgrade_backup',
    route = f'/cmapi/{_version}/cluster/preupgrade-backup',
    action = 'preupgrade_backup',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/upgrade-mdb-mcs (PUT)
dispatcher.connect(
    name = 'node_upgrade_mdb_mcs',
    route = f'/cmapi/{_version}/node/upgrade-mdb-mcs',
    action = 'upgrade_mdb_mcs',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/upgrade-mdb-mcs (PUT)
dispatcher.connect(
    name = 'cluster_upgrade_mdb_mcs',
    route = f'/cmapi/{_version}/cluster/upgrade-mdb-mcs',
    action = 'upgrade_mdb_mcs',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/kick-cmapi-upgrade (PUT)
dispatcher.connect(
    name = 'node_kick_cmapi_upgrade',
    route = f'/cmapi/{_version}/node/kick-cmapi-upgrade',
    action = 'kick_cmapi_upgrade',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)


# /_version/cluster/upgrade-cmapi (PUT)
dispatcher.connect(
    name = 'cluster_upgrade_cmapi',
    route = f'/cmapi/{_version}/cluster/upgrade-cmapi',
    action = 'upgrade_cmapi',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/check-shared-file/ (GET)
dispatcher.connect(
    name = 'node_check_shared_file',
    route = f'/cmapi/{_version}/node/check-shared-file',
    action = 'check_shared_file',
    controller = NodeController(),
    conditions = {'method': ['GET']}
)


# /_version/cluster/check-shared-storage/ (PUT)
dispatcher.connect(
    name = 'cluster_check_shared_storage',
    route = f'/cmapi/{_version}/cluster/check-shared-storage',
    action = 'check_shared_storage',
    controller = ClusterController(),
    conditions = {'method': ['PUT']}
)


# /_version/node/stateful-config/ (PUT)
dispatcher.connect(
    name = 'node_put_stateful_config',
    route = f'/cmapi/{_version}/node/stateful-config',
    action = 'put_stateful_config',
    controller = NodeController(),
    conditions = {'method': ['PUT']}
)

# /_version/cmapi_config (PATCH)
dispatcher.connect(
    name = 'cmapi_config',
    route = f'/cmapi/{_version}/cmapi_config',
    action = 'patch_cmapi_config',
    controller = CmapiConfigController(),
    conditions = {'method': ['PATCH']}
)

def jsonify_error(status, message, traceback, version):
    # pylint: disable=unused-argument
    """JSONify all CherryPy error responses (created by raising the
    cherrypy.HTTPError exception)
    """

    cherrypy.response.headers['Content-Type'] = 'application/json'
    response_body = json.dumps(
        {
            'error': {
                'http_status': status,
                'message': message,
            }
        }
    )

    cherrypy.response.status = status

    return response_body


def jsonify_404(status, message, traceback, version):
    # pylint: disable=unused-argument
    """Specialized renderer for 404 Not Found that logs context, then renders JSON.
    """
    try:
        req = cherrypy.request
        method = getattr(req, 'method', '')
        path = getattr(req, 'path_info', '') or '/'
        remote_ip = getattr(getattr(req, 'remote', None), 'ip', '') or '?'
        logger.error("404 Not Found: %s %s from %s", method, path, remote_ip)
    except Exception:
        pass
    return jsonify_error(status, message, traceback, version)



import logging
import contextvars
from typing import Any, Dict, Optional

import sentry_sdk
from sentry_sdk.tracing import Transaction

from tracing.tracer import TraceSpan, TracerBackend
from tracing.utils import swallow_exceptions


logger = logging.getLogger(__name__)


class SentryBackend(TracerBackend):
    """Mirror spans and events from our Tracer into Sentry SDK."""

    def __init__(self) -> None:
        self._active_spans: Dict[str, Any] = {}
        self._current_transaction = contextvars.ContextVar[Optional[Transaction]]("sentry_transaction", default=None)

    @swallow_exceptions
    def on_span_start(self, span: TraceSpan) -> None:
        kind_to_op = {
            'SERVER': 'http.server',
            'CLIENT': 'http.client',
            'INTERNAL': 'internal',
        }
        op = kind_to_op.get(span.kind.upper(), 'internal')
        sdk_span = sentry_sdk.start_span(op=op, description=span.name)
        sdk_span.set_tag('w3c.trace_id', span.trace_id)
        sdk_span.set_tag('w3c.span_id', span.span_id)
        if span.parent_span_id:
            sdk_span.set_tag('w3c.parent_span_id', span.parent_span_id)
        if span.attributes:
            sdk_span.set_data('cmapi.span_attributes', dict(span.attributes))
        sdk_span.__enter__()
        self._active_spans[span.span_id] = sdk_span

    @swallow_exceptions
    def on_span_end(self, span: TraceSpan, exc: Optional[BaseException]) -> None:
        sdk_span = self._active_spans.pop(span.span_id, None)
        if sdk_span is None:
            return
        if exc is not None:
            sdk_span.set_status('internal_error')
        sdk_span.__exit__(
            type(exc) if exc else None,
            exc,
            exc.__traceback__ if exc else None
        )

    @swallow_exceptions
    def on_span_event(self, span: TraceSpan, name: str, attrs: Dict[str, Any]) -> None:
        sentry_sdk.add_breadcrumb(category='event', message=name, data=dict(attrs))

    @swallow_exceptions
    def on_span_status(self, span: TraceSpan, code: str, description: str) -> None:
        sdk_span = self._active_spans.get(span.span_id)
        if sdk_span is not None:
            sdk_span.set_status(code)

    @swallow_exceptions
    def on_span_exception(self, span: TraceSpan, exc: BaseException) -> None:
        sentry_sdk.capture_exception(exc)

    @swallow_exceptions
    def on_span_attribute(self, span: TraceSpan, key: str, value: Any) -> None:
        sdk_span = self._active_spans.get(span.span_id)
        if sdk_span is not None:
            sdk_span.set_data(f'attr.{key}', value)

    @swallow_exceptions
    def on_inject_headers(self, headers: Dict[str, str]) -> None:
        traceparent = sentry_sdk.get_traceparent()
        baggage = sentry_sdk.get_baggage()
        if traceparent:
            headers['sentry-trace'] = traceparent
        if baggage:
            headers['baggage'] = baggage

    @swallow_exceptions
    def on_incoming_request(self, headers: Dict[str, str], method: str, path: str) -> None:
        name = f"{method} {path}" if method or path else "http.server"
        # Continue from incoming headers, then START the transaction per SDK v2
        continued = sentry_sdk.continue_trace(headers or {}, op='http.server', name=name)
        transaction = sentry_sdk.start_transaction(transaction=continued)
        # Store started transaction in context var and make current (enter)
        self._current_transaction.set(transaction)
        transaction.__enter__()
        scope = sentry_sdk.Hub.current.scope
        if method:
            scope.set_tag('http.method', method)
        if path:
            scope.set_tag('http.path', path)

    @swallow_exceptions
    def on_request_finished(self, status_code: Optional[int]) -> None:
        transaction = self._current_transaction.get()
        if transaction is None:
            return
        if status_code is not None:
            transaction.set_http_status(status_code)
        # Exit to restore parent and finish the transaction
        transaction.__exit__(None, None, None)
        # Clear transaction in this context
        self._current_transaction.set(None)





import logging
import os
import subprocess
import unittest
import xml.etree.ElementTree as ET
from pathlib import Path
from shutil import copyfile
from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from lxml import etree

from cmapi_server.constants import CMAPI_DEFAULT_CONF_PATH
from mcs_node_control.models.dbrm import (
    DBRM, set_cluster_mode
)
from mcs_node_control.models.node_config import NodeConfig
from mcs_node_control.models.misc import read_module_id
from mcs_node_control.models.node_status import NodeStatus
from mcs_node_control.test.settings import CONFIG_PATH_NEW, CONFIG_PATH_OLD


MCS_NODE_MODELS = 'mcs_node_control.models'
NODE_CONFIG_MODULE = f'{MCS_NODE_MODELS}.node_config'


logging.basicConfig(level='DEBUG')


# These tests needs working DBRM worker.
class NodeConfigTest(TestCase):

    @mock.patch(f'{NODE_CONFIG_MODULE}.mkdir')
    @mock.patch(f'{NODE_CONFIG_MODULE}.chown')
    @mock.patch(f'{NODE_CONFIG_MODULE}.read_module_id', return_value=1)
    @mock.patch(
        f'{NODE_CONFIG_MODULE}.NodeConfig.in_active_nodes',
        return_value=False
    )
    def test_apply_config(self, *_args):
        """Test apply configuration file."""
        with TemporaryDirectory() as tmp_dirname:
            config_filepath = os.path.join(tmp_dirname, 'Columnstore.xml')

            copyfile(CONFIG_PATH_OLD, config_filepath)
            # change config
            parser = etree.XMLParser(load_dtd=True)
            # new_tree = etree.parse('/etc/columnstore/Columnstore.xml', parser=parser)
            new_tree = etree.parse(CONFIG_PATH_NEW, parser=parser)

            node_config = NodeConfig()
            xml_string = node_config.to_string(new_tree)

            node_config.apply_config(config_filepath, xml_string)

            # compare configurations
            config_file = Path(config_filepath)
            xml_string_written = config_file.read_text()
            self.assertEqual(xml_string_written, xml_string)
            # copy must exists
            config_file_copy = Path(f"{config_filepath}.cmapi.save")
            self.assertTrue(config_file_copy.exists())

    @mock.patch(f'{NODE_CONFIG_MODULE}.mkdir')
    @mock.patch(f'{NODE_CONFIG_MODULE}.chown')
    @mock.patch(f'{NODE_CONFIG_MODULE}.read_module_id', return_value=1)
    @mock.patch(
        f'{NODE_CONFIG_MODULE}.NodeConfig.in_active_nodes',
        return_value=False
    )
    def test_rollback_config(self, *_args):
        """"Test rollback applied configuration file."""
        with TemporaryDirectory() as tmp_dirname:
            config_filepath = os.path.join(tmp_dirname, 'Columnstore.xml')
            copyfile(CONFIG_PATH_OLD, config_filepath)

            old_config_file = Path(CONFIG_PATH_OLD)
            old_xml_string = old_config_file.read_text()
            new_config_file = Path(CONFIG_PATH_NEW)
            new_xml_string = new_config_file.read_text()

            node_config = NodeConfig()
            node_config.apply_config(config_filepath, new_xml_string)
            node_config.rollback_config(config_filepath)

            config_file = Path(config_filepath)
            xml_string_restored = config_file.read_text()
            self.assertEqual(xml_string_restored, old_xml_string)

    def test_get_current_config(self):
        """Test get current config from file."""
        config_file = Path(CONFIG_PATH_OLD)
        node_config = NodeConfig()
        self.assertEqual(
            node_config.get_current_config(CONFIG_PATH_OLD),
            config_file.read_text()
        )

    def test_set_cluster_mode(self):
        """Test set cluster mode.

        TODO:
            - move from here. There are no set_cluster_mode in NodeConfig
            - split to unit and integrational tests
            - make unittests for raising exception
        """

        for mode in ['readonly', 'readwrite']:
            with self.subTest(mode=mode):
                fake_mode = mode
                set_cluster_mode(mode)
                with DBRM() as dbrm:
                    if dbrm.get_dbrm_status() != 'master':
                        fake_mode = 'readonly'
                    self.assertEqual(dbrm.get_cluster_mode(), fake_mode)
                    self.assertEqual(dbrm._get_cluster_mode(), mode)

    def test_get_dbrm_conn_info(self):
        node_config = NodeConfig()
        root = node_config.get_current_config_root(CONFIG_PATH_OLD)
        master_conn_info = node_config.get_dbrm_conn_info(root)

        tree = ET.parse(CONFIG_PATH_OLD)
        master_ip = tree.find('./DBRM_Controller/IPAddr').text
        master_port = tree.find('./DBRM_Controller/Port').text

        self.assertEqual(master_conn_info['IPAddr'], master_ip)
        self.assertEqual(master_conn_info['Port'], master_port)

    def test_is_primary_node(self):
        try:
            current_master = None
            node_config = NodeConfig()
            root = node_config.get_current_config_root()
            current_master = node_config.get_dbrm_conn_info(root)['IPAddr']
            list_ips = "ip -4 -o addr | awk '!/^[0-9]*: ?lo|link\/ether/ {print $4}'"
            result = subprocess.run(list_ips,
                                     shell=True,
                                     stdout=subprocess.PIPE)
            local_addresses = result.stdout.decode('ASCII').split('\n')
            local_addresses = [addr.split('/')[0] for addr in local_addresses if len(addr)]
            os.system(f"mcsSetConfig DBRM_Controller IPAddr {local_addresses[0]}")
            self.assertTrue(node_config.is_primary_node())
            os.system(f"mcsSetConfig DBRM_Controller IPAddr 8.8.8.8")
            self.assertFalse(node_config.is_primary_node())
            os.system(f"mcsSetConfig DBRM_Controller IPAddr {current_master}")
        except AssertionError as e:
            if current_master is not None:
                os.system(f"mcsSetConfig DBRM_Controller IPAddr \
{current_master}")
            raise e

    def test_get_network_interfaces(self):
        node_config = NodeConfig()
        addresses = list(node_config.get_network_addresses())
        exemplar_addresses = []
        list_ips = "ip -4 -o addr | awk '!/^[0-9]*: ?lo|link\/ether/ {print $4}'"
        result = subprocess.run(list_ips,
                                 shell=True,
                                 stdout=subprocess.PIPE)
        exemplar_addresses += result.stdout.decode('ASCII').split('\n')
        list_ips = "ip -6 -o addr | awk '!/^[0-9]*: ?lo|link\/ether/ {print $4}'"
        result = subprocess.run(list_ips,
                                 shell=True,
                                 stdout=subprocess.PIPE)
        exemplar_addresses += result.stdout.decode('ASCII').split('\n')
        golden_addresses = [addr.split('/')[0] for addr in exemplar_addresses if len(addr) > 0]
        for addr in golden_addresses:
            self.assertTrue(addr in addresses)

    def test_is_single_node(self):
        try:
            current_master = None
            node_config = NodeConfig()
            root = node_config.get_current_config_root()
            current_master = node_config.get_dbrm_conn_info(root)['IPAddr']
            os.system(f"mcsSetConfig DBRM_Controller IPAddr 127.0.0.1")
            self.assertTrue(node_config.is_single_node())
            os.system(f"mcsSetConfig DBRM_Controller IPAddr 8.8.8.8")
            self.assertFalse(node_config.is_single_node())
            os.system(f"mcsSetConfig DBRM_Controller IPAddr {current_master}")
        except AssertionError as e:
            if current_master is not None:
                os.system(f"mcsSetConfig DBRM_Controller IPAddr \
{current_master}")
            raise e

    @mock.patch(f'{NODE_CONFIG_MODULE}.read_module_id', return_value=1)
    def test_get_module_net_address(self, *args):
        with TemporaryDirectory() as tmp_dirname:
            config_filepath = os.path.join(tmp_dirname, 'Columnstore.xml')
            copyfile(CONFIG_PATH_OLD, config_filepath)

            module_address = None
            node_config = NodeConfig()
            current_module_id = read_module_id()
            module_address_sh = (
                f'mcsGetConfig -c {config_filepath} '
                f'SystemModuleConfig ModuleIPAddr{current_module_id}-1-3'
            )
            result = subprocess.run(
                module_address_sh, shell=True, stdout=subprocess.PIPE
            )
            module_address = result.stdout.decode('ASCII').split('\n')[0]
            dummy_address = '8.8.8.8'
            os.system(
                f'mcsSetConfig -c {config_filepath} '
                f'SystemModuleConfig ModuleIPAddr{current_module_id}-1-3 '
                f'{dummy_address}'
            )
            root = node_config.get_current_config_root(config_filepath)
            self.assertEqual(
                dummy_address, node_config.get_module_net_address(root)
            )
            self.assertNotEqual(
                module_address, node_config.get_module_net_address(root)
            )
            os.system(
                f'mcsSetConfig -c {config_filepath} SystemModuleConfig '
                f'ModuleIPAddr{current_module_id}-1-3 {module_address}'
            )
            root = node_config.get_current_config_root(config_filepath)
            self.assertEqual(
                module_address, node_config.get_module_net_address(root)
            )

    def test_get_new_module_id(self):
        try:
            current_module_id = None
            current_module_address = None
            node_config = NodeConfig()
            current_module_id = read_module_id()
            root = node_config.get_current_config_root()
            current_module_address = node_config.get_module_net_address(root)
            os.system(f"mcsSetConfig SystemModuleConfig \
    ModuleIPAddr{current_module_id}-1-3 8.8.8.8")
            os.system(f"mcsSetConfig SystemModuleConfig \
    ModuleIPAddr{current_module_id+42}-1-3 {current_module_address}")
            root = node_config.get_current_config_root()
            self.assertEqual(current_module_id+42,
                             node_config.get_new_module_id(root))
            self.assertNotEqual(current_module_id,
                                node_config.get_new_module_id(root))
            os.system(f"mcsSetConfig SystemModuleConfig \
    ModuleIPAddr{current_module_id}-1-3 {current_module_address}")
            os.system(f"mcsSetConfig -x SystemModuleConfig \
    ModuleIPAddr{current_module_id+42}-1-3 {current_module_address}")
            root = node_config.get_current_config_root()
            self.assertEqual(current_module_id,
                         node_config.get_new_module_id(root))
        except AssertionError as e:
            if current_module_id is not None and current_module_address is not None:
                os.system(f"mcsSetConfig SystemModuleConfig \
        ModuleIPAddr{current_module_id}-1-3 {current_module_address}")
                os.system(f"mcsSetConfig -x SystemModuleConfig \
    ModuleIPAddr{current_module_id+42}-1-3 {current_module_address}")

    def test_dbroots_to_create(self):
        try:
            node_config = NodeConfig()
            current_module_id = read_module_id()
            dummy_dbroots = [42, 43]
            dbroot_seq_id = 2
            for d in dummy_dbroots:
                os.system(f"mcsSetConfig SystemModuleConfig \
    ModuleDBRootID{current_module_id}-{dbroot_seq_id}-3 {d}")
                dbroot_seq_id += 1
            root = node_config.get_current_config_root()
            dbroots_to_create = list(node_config.dbroots_to_create(root=root,                                                                                                         module_id=current_module_id))
            for d in dbroots_to_create:
                self.assertTrue(d in dummy_dbroots)
        except AssertionError as e:
            dbroot_seq_id = 2
            for d in dummy_dbroots:
                os.system(f"mcsSetConfig -x SystemModuleConfig \
    ModuleDBRootID{current_module_id}-{dbroot_seq_id}-3 {d}")
                dbroot_seq_id += 1
            raise e

        dbroot_seq_id = 2
        for d in dummy_dbroots:
            os.system(f"mcsSetConfig -x SystemModuleConfig \
ModuleDBRootID{current_module_id}-{dbroot_seq_id}-3 {d}")
            dbroot_seq_id += 1


if __name__ == '__main__':
    unittest.main()



import io
import logging
import unittest

from mcs_node_control.models.dbrm_socket import MAGIC_BYTES, DBRMSocketHandler


logging.basicConfig(level='DEBUG')


class TestDBRMSocketHandler(unittest.TestCase):

    def test_myreceive_to_magic(self):
        response_data = b'\x01\x00\x00\x00\x00'
        valid_magic = b'%s%s' % (MAGIC_BYTES, response_data)
        first_unknow = b'A%s%s' % (MAGIC_BYTES, response_data)
        partial_first_magic = b'%s%s%s' % (
            MAGIC_BYTES[:3], MAGIC_BYTES, response_data
        )
        sock_responses = [valid_magic, first_unknow, partial_first_magic]
        for sock_response in sock_responses:
            with self.subTest(sock_response=sock_response):
                data_stream = io.BytesIO(sock_response)
                data_stream.recv = data_stream.read
                dbrm_socket = DBRMSocketHandler()
                # pylint: disable=protected-access
                dbrm_socket._socket = data_stream
                dbrm_socket._receive_magic()
                self.assertEqual(data_stream.read(), response_data)



import os
import time

import psutil

PROCFS_PATH = '/proc/' # Linux only


def open_binary(fname, **kwargs):
    return open(fname, "rb", **kwargs)


def get_host_uptime():
    """
    Return the system boot time expressed in seconds since the epoch.

    :rtype: int : diff b/w current epoch and boot epoch
    """
    path = f'{PROCFS_PATH}stat'
    boot_time = 0
    with open_binary(path) as f:
        for line in f:
            if line.startswith(b'btime'):
                boot_time = float(line.strip().split()[1])
                return int(time.time() - int(boot_time))
        return 0


class Process():
    """An interface to retrieve data from proc."""
    def get_proc_iterator(self):
        for pid in self.pids():
            yield pid


    def pids(self):
        """Returns a list of PIDs currently running on the system."""
        return [int(x) for x in os.listdir(PROCFS_PATH) if x.isdigit()]


    def name(self, pid: int):
        """Method to retrive name associated with the pid."""
        return self.parse_stat_file(pid)['name']


    def parse_stat_file(self, pid: int):
        """Parse /proc/{pid}/stat file and return a dict with various
        process info.

        Using "man proc" as a reference: where "man proc" refers to
        position N always substract 3 (e.g ppid position 4 in
        'man proc' == position 1 in here).
        """
        ret = {}
        try:
            with open_binary(f"{PROCFS_PATH}{pid}/stat") as f:
                data = f.read()
            # Process name is between parentheses. It can contain spaces and
            # other parentheses. This is taken into account by looking for
            # the first occurrence of "(" and the last occurence of ")".
            rpar = data.rfind(b')')
            name = data[data.find(b'(') + 1:rpar]
            fields = data[rpar + 2:].split()

            ret['name'] = name
            ret['status'] = fields[0]
            ret['ppid'] = fields[1]
            ret['ttynr'] = fields[4]
            ret['utime'] = fields[11]
            ret['stime'] = fields[12]
            ret['children_utime'] = fields[13]
            ret['children_stime'] = fields[14]
            ret['create_time'] = fields[19]
            ret['cpu_num'] = fields[36]
            ret['blkio_ticks'] = fields[39]  # aka 'delayacct_blkio_ticks'
        except (PermissionError, ProcessLookupError, FileNotFoundError):
            ret['name'] = ''
            ret['status'] = ''
            ret['ppid'] = ''
            ret['ttynr'] = ''
            ret['utime'] = ''
            ret['stime'] = ''
            ret['children_utime'] = ''
            ret['children_stime'] = ''
            ret['create_time'] = ''
            ret['cpu_num'] = ''
            ret['blkio_ticks'] = ''  # aka 'delayacct_blkio_ticks'
        return ret

    @staticmethod
    def check_process_alive(proc_name: str) -> bool:
        """Check process running.

        :param proc_name: process name
        :type proc_name: str
        :return: True if process running, otherwise False
        :rtype: bool
        """
        # Iterate over the all the running process
        for proc in psutil.process_iter():
            try:
                # Check if process name equals to the given name string.
                if proc_name.lower() == proc.name().lower():
                    return True
            except (
                psutil.NoSuchProcess, psutil.AccessDenied,
                psutil.ZombieProcess
            ):
                pass
        return False



import logging
import time
import threading

from .heartbeater import HeartBeater
from .config import Config
from .heartbeat_history import HBHistory
from .agent_comm import AgentComm
from .shared_storage_monitor import SharedStorageMonitor

from cmapi_server.managers.application import AppStatefulConfig


class NodeMonitor:

    def __init__(
        self, agent=None, config=None, samplingInterval=30,
        flakyNodeThreshold=0.5
    ):
        self._agentComm = AgentComm(agent)
        self._die = False
        self._inStandby = False
        self._testMode = False  # TODO: remove
        self._hbHistory = HBHistory()
        self._logger = logging.getLogger('node_monitor')
        self._runner = None
        if config is not None:
            self._config = config
        else:
            self._config = Config()
        self._hb = HeartBeater(self._config, self._hbHistory)
        self.samplingInterval = samplingInterval
        # not used yet, KI-V-SS for V1 [old comment from Patrick]
        self.flakyNodeThreshold = flakyNodeThreshold
        self.myName = self._config.who_am_I()
        # Pass heartbeat history to shared storage monitor so it can skip
        # nodes that just dropped from Good to NoResponse.
        self.shared_storage_monitor = SharedStorageMonitor(hb_history=self._hbHistory)
        #self._logger.info("Using {} as my name".format(self.myName))

    def __del__(self):
        self.stop()

    def start(self):
        self._agentComm.start()
        self._hb.start()
        self.shared_storage_monitor.start()
        self._die = False
        self._runner = threading.Thread(
            target=self.monitor, name='NodeMonitor'
        )
        self._runner.start()

    def stop(self):
        self._die = True
        self._agentComm.die()
        if not self._testMode:
            self._hb.stop()
        self._runner.join()
        self.shared_storage_monitor.stop()

    def _removeRemovedNodes(self, desiredNodes):
        self._hbHistory.keepOnlyTheseNodes(desiredNodes)

    def _pickNewActor(self, nodes):
        if not nodes:
            return
        if self.myName == nodes[0]:
            self._isActorOfCohort = True
        else:
            self._isActorOfCohort = False

    def _chooseNewPrimaryNode(self):
        self._agentComm.movePrimaryNode()

    def monitor(self):
        while not self._die:
            try:
                self._logger.info('Starting the monitor logic')
                self._monitor()
            except Exception:
                self._logger.error(
                    f'monitor() caught an exception.',
                    exc_info=True
                )
            if not self._die:
                time.sleep(1)
        self._logger.info("node monitor logic exiting normally...")

    def _monitor(self):
        """
        This works like the main loop of a game.
        1) check current state
        2) identify the differences
        3) update based on the differences
        """

        (desiredNodes, activeNodes, inactiveNodes) = self._config.getAllNodes()
        self._pickNewActor(activeNodes)

        logged_idleness_msg = False
        logged_active_msg = False
        inStandbyMode = False
        while not self._die:
            # these things would normally go at the end of the loop; doing it here
            # to reduce line count & chance of missing something as we add more code
            oldActiveNodes = activeNodes
            wasActorOfCohort = self._isActorOfCohort
            self._logger.trace(
                f'Previous actor of cohort state is {wasActorOfCohort}'
            )
            time.sleep(1)

            # get config updates
            (desiredNodes, activeNodes, inactiveNodes) = self._config.getAllNodes()
            self.myName = self._config.who_am_I()
            self.primaryNode = self._config.getPrimaryNode()

            # remove nodes from history that have been removed from the cluster
            self._removeRemovedNodes(desiredNodes)

            # if there are less than 3 nodes in the cluster or shared storage is OFF,
            # failover must not perform any actions. Keep the thread alive but idle,
            # so that when conditions change (e.g., shared storage becomes available),
            # the monitor can resume work without restart.
            if len(desiredNodes) < 3 or not AppStatefulConfig.is_shared_storage():
                if not logged_idleness_msg:
                    self._logger.info(
                        'Failover support is inactive; '
                        'requires at least 3 nodes and a shared storage system'
                    )
                    logged_idleness_msg = True
                    logged_active_msg = False
                # skip the rest of the loop to avoid sending heartbeats or performing any failover
                # actions while inactive.
                continue
            elif not logged_active_msg:
                self._logger.info(
                    'Failover support is active, '
                    f'monitoring nodes {desiredNodes}'
                )
                logged_active_msg = True
                logged_idleness_msg = False

            # nothing to do in this case
            if len(desiredNodes) == 1:
                continue

            # has this node been reactivated?
            if self.myName in activeNodes:
                #TODO: remove useless flag or use it in future releases
                self._inStandby = False
            # has it been deactivated?
            else:
                self._logger.trace('Node not in active nodes, do nothing.')
                self._inStandby = True
                continue    # wait to be activated

            # send heartbeats
            self._hb.sendHeartbeats()

            # decide if action is necessary based on config changes

            # get the list of nodes no longer responding to heartbeats
            # V1: only remove a node that hasn't responded to any pings in the sampling period
            deactivateSet = set()
            for node in activeNodes:
                if node == self.myName:
                    continue
                history = self._hbHistory.getNodeHistory(node, self.samplingInterval, HBHistory.GoodResponse)
                self._logger.trace(f'Get history "{history}" for node {node}')
                noResponses = [ x for x in history if x == HBHistory.NoResponse ]
                if len(noResponses) == self.samplingInterval:
                    deactivateSet.add(node)

            # get the list of nodes that have started responding
            # reactivate live nodes that have begun responding to heartbeats
            # V1: only reactivate a node if we have good responses for the whole sampling period
            activateSet = set()
            for node in inactiveNodes:
                history = self._hbHistory.getNodeHistory(node, self.samplingInterval, HBHistory.NoResponse)
                goodResponses = [ x for x in history if x == HBHistory.GoodResponse ]
                if len(goodResponses) == self.samplingInterval:
                    activateSet.add(node)

            # effectiveActiveNodeList can be described as activeNodes after pending config changes
            # have been applied.  Another way to view it is that it reflects current reality, whereas
            # the config file reflects a fixed point in the recent past.
            effectiveActiveNodeList = sorted((set(activeNodes) - deactivateSet) | activateSet)

            # if there was a change to the list of active nodes
            # decide if this node is the effective actor in the cohort.
            if effectiveActiveNodeList != activeNodes:
                self._pickNewActor(effectiveActiveNodeList)
                self._logger.trace(
                    f'Effective list changed, actor state is {self._isActorOfCohort}'
                )
            elif oldActiveNodes != activeNodes:
                self._pickNewActor(activeNodes)
                self._logger.trace(
                    f'Active list changed, actor state is {self._isActorOfCohort}'
                )


            # if we are in a cohort that has <= 50% of the desired nodes, enter standby
            if len(activeNodes)/len(desiredNodes) <= 0.5 and len(effectiveActiveNodeList)/len(desiredNodes) <= 0.5:
                if not inStandbyMode:
                    msg = "Only {} out of {} nodes are active.  At least {} are required.  Entering standby mode to protect the system."\
                        .format(len(activeNodes), len(desiredNodes), int(len(desiredNodes)/2) + 1)
                    self._agentComm.raiseAlarm(msg)
                    self._logger.critical(msg)
                    self._agentComm.enterStandbyMode()
                    inStandbyMode = True
                continue
            elif inStandbyMode and len(effectiveActiveNodeList)/len(desiredNodes) > 0.5:
                self._logger.info("Exiting standby mode, waiting for config update")
                inStandbyMode = False

            # (wasActorOfCohort and not isActorOfCohort) indicates that a new Actor has come online.
            # To hand over the crown, perform one last act as Actor to add it back to the cluster
            # and synchronize its config file.

            # if not the actor, nothing else for this node to do
            if not self._isActorOfCohort and not wasActorOfCohort:
                continue

            # as of here, this node is the actor of its quorum

            if len(deactivateSet) > 0:
                self._agentComm.deactivateNodes(list(deactivateSet))

            if len(activateSet) > 0:
                self._agentComm.activateNodes(activateSet)

            # if the primary node is in this list to be deactivated, or its already on the inactive list
            # choose a new primary node.  The deadNode list is a sanity check for cases like the cluster
            # starting with the primary node already in inactive-nodes.
            deadNodeList = list(deactivateSet) + inactiveNodes
            if self.primaryNode in deadNodeList:
                self._chooseNewPrimaryNode()

    # methods for testing
    def turnOffHBResponder(self):
        self.stop()



import os

from cmapi_server.constants import MCS_INSTALL_BIN


MCS_CLI_ROOT_PATH = os.path.dirname(__file__)
MCS_CLI_LOG_CONF_PATH =  os.path.join(MCS_CLI_ROOT_PATH, 'mcs_cli_log.conf')

MCS_COLUMNSTORE_REVIEW_SH = os.path.join(
    MCS_INSTALL_BIN, 'columnstore_review.sh'
)

INSTALL_ES_LOG_FILEPATH = '/var/tmp/mcs_cli_install_es.log'



import unittest
from unittest.mock import patch, mock_open, MagicMock
import socket
import ssl
import json
import os
import subprocess
from urllib.error import HTTPError, URLError
from urllib.request import Request
from xml.etree.ElementTree import Element
import importlib.util

import mcssavebrm

class TestMcsSavebrmFunctions(unittest.TestCase):
    @patch('mcssavebrm.configparser.ConfigParser.get', return_value='test_api_key')
    @patch('mcssavebrm.configparser.ConfigParser.read')
    def test_get_api_key(self, mock_read, mock_get):
        self.assertEqual(mcssavebrm.get_api_key(), 'test_api_key')
        mock_read.assert_called_once_with(mcssavebrm.CMAPI_CONFIG_PATH)

    def test_get_unverified_context(self):
        ctx = mcssavebrm.get_unverified_context()
        self.assertFalse(ctx.check_hostname)
        self.assertEqual(ctx.verify_mode, ssl.CERT_NONE)

    @patch('mcssavebrm.urlopen')
    @patch('mcssavebrm.get_unverified_context')
    def test_cmapi_available(self, mock_get_unverified_context, mock_urlopen):
        mock_get_unverified_context.return_value = ssl._create_unverified_context()
        mock_urlopen.side_effect = HTTPError(None, 404, 'Not Found', None, None)
        self.assertTrue(mcssavebrm.cmapi_available())

    @patch('mcssavebrm.fcntl.ioctl')
    @patch('mcssavebrm.socket.socket')
    def test_get_ip_address_by_nic(self, mock_socket, mock_ioctl):
        mock_socket_inst = MagicMock()
        mock_socket.return_value = mock_socket_inst
        mock_ioctl.return_value = b'\x00' * 20 + b'\x7f\x00\x00\x01'
        self.assertEqual(mcssavebrm.get_ip_address_by_nic('lo'), '127.0.0.1')

    @patch('mcssavebrm.get_ip_address_by_nic', return_value='127.0.0.1')
    @patch('mcssavebrm.socket.gethostbyaddr', return_value=('localhost', [], []))
    @patch('mcssavebrm.socket.if_nameindex', return_value=[(1, 'lo')])
    def test_is_primary_fallback(self, mock_if_nameindex, mock_gethostbyaddr, mock_get_ip_address_by_nic):
        self.assertTrue(mcssavebrm.is_primary_fallback('localhost'))

    @patch('mcssavebrm.cmapi_available', return_value=True)
    @patch('mcssavebrm.urlopen')
    @patch('mcssavebrm.get_unverified_context')
    @patch('mcssavebrm.get_api_key', return_value='test_api_key')
    def test_is_node_primary(self, mock_get_api_key, mock_get_unverified_context, mock_urlopen, mock_cmapi_available):
        mock_get_unverified_context.return_value = ssl._create_unverified_context()
        mock_response = MagicMock()
        mock_response.read.return_value = json.dumps({'is_primary': 'True'}).encode('utf-8')
        mock_urlopen.return_value.__enter__.return_value = mock_response
        root = MagicMock()
        self.assertTrue(mcssavebrm.is_node_primary(root))

    @patch('os.path.getsize', return_value=1024)
    def test_get_file_size(self, mock_getsize):
        self.assertEqual(mcssavebrm.get_file_size('test_file'), 1024)

    @patch('mcssavebrm.get_file_size', return_value=500)
    def test_em_is_empty(self, mock_get_file_size):
        self.assertTrue(mcssavebrm.em_is_empty('test_prefix'))

    @patch('os.remove')
    @patch('os.listdir', return_value=[f'test_file{i}' for i in range(50)])
    def test_clean_up_backup_brm_files(self, mock_listdir, mock_remove):
        mcssavebrm.clean_up_backup_brm_files('/dummy/path')
        self.assertEqual(mock_remove.call_count, 10)

    @patch('os.remove')
    @patch('mcssavebrm.glob.glob', return_value=['test_file1', 'test_file2'])
    def test_remove_files_by_prefix_if_exist(self, mock_glob, mock_remove):
        mcssavebrm.remove_files_by_prefix_if_exist('test_prefix')
        mock_remove.assert_any_call('test_file1')
        mock_remove.assert_any_call('test_file2')

    @patch('xml.etree.ElementTree.parse')
    def test_get_config_root_from_file(self, mock_parse):
        mock_tree = MagicMock()
        mock_parse.return_value = mock_tree
        self.assertEqual(mcssavebrm.get_config_root_from_file('test_file'), mock_tree.getroot())

    @patch('time.time', return_value=1624478400)
    def test_get_epoch_prefix(self, mock_time):
        self.assertEqual(mcssavebrm.get_epoch_prefix(), 'backup_1624478400')

    @patch('mcssavebrm.get_epoch_prefix', return_value='backup_1624478400')
    @patch('mcssavebrm.get_save_brm_dir_path', return_value='/tmp/columnstore_tmp_files/rdwrscratch/')
    def test_get_save_brm_path_prefix(self, mock_get_epoch_prefix, mock_get_save_brm_dir_path):
        root = MagicMock()
        self.assertIn('backup_1624478400_BRM_saves', mcssavebrm.get_save_brm_path_prefix(root))

    @patch('subprocess.check_call')
    def test_call_save_brm(self, mock_check_call):
        self.assertEqual(mcssavebrm.call_save_brm('test_path'), 'test_path')

    @patch('mcssavebrm.call_save_brm', return_value='test_path')
    @patch('mcssavebrm.get_save_brm_path_prefix', return_value='test_path')
    def test_call_save_brm_locally(self, mock_get_save_brm_path_prefix, mock_call_save_brm):
        root = MagicMock()
        self.assertEqual(mcssavebrm.call_save_brm_locally(root), 'test_path')

    @patch('subprocess.check_call')
    @patch('mcssavebrm.get_save_brm_path_prefix', return_value='test_path')
    def test_call_save_brm_with_local_fallback(self, mock_get_save_brm_path_prefix, mock_check_call):
        root = MagicMock()
        mock_check_call.side_effect = [subprocess.CalledProcessError(1, 'test'), None]
        with self.assertRaises(SystemExit):
            mcssavebrm.call_save_brm_with_local_fallback(root)


if __name__ == '__main__':
    unittest.main()


import json
import cherrypy as cp 

class APIError(cp.HTTPError):
    def __init__(self, status: int = 500, message: str = ''):
        super().__init__(status=status)
        self._error_message = message

    def set_response(self):
        super().set_response()
        response = cp.serving.response
        response.body = json.dumps({'error': self._error_message}).encode()



import logging
import threading
from os.path import getmtime

from cmapi_server.constants import DEFAULT_FAILOVER_SAMPLING_INTERVAL_SECS, DEFAULT_MCS_CONF_PATH
from mcs_node_control.models.node_config import NodeConfig


class Config:
    config_file = ''

    # params read from the config file
    _desired_nodes = []
    _active_nodes = []
    _inactive_nodes = []
    _primary_node = ''
    _my_name = None  # derived from config file
    _failover_sampling_interval = DEFAULT_FAILOVER_SAMPLING_INTERVAL_SECS

    config_lock = threading.Lock()
    last_mtime = 0
    die = False
    logger = None

    def __init__(self, config_file=DEFAULT_MCS_CONF_PATH):
        self.config_file = config_file
        self.logger = logging.getLogger()

    def getDesiredNodes(self):
        self.config_lock.acquire()
        self.check_reload()
        ret = self._desired_nodes
        self.config_lock.release()
        return ret

    def getActiveNodes(self):
        self.config_lock.acquire()
        self.check_reload()
        ret = self._active_nodes
        self.config_lock.release()
        return ret

    def getInactiveNodes(self):
        self.config_lock.acquire()
        self.check_reload()
        ret = self._inactive_nodes
        self.config_lock.release()
        return ret

    def getAllNodes(self):
        """Returns a 3-element tuple describing the status of all nodes.

        index 0 = all nodes in the cluster
        index 1 = all active nodes
        index 2 = all inactive nodes
        """
        self.config_lock.acquire()
        self.check_reload()
        ret = (self._desired_nodes, self._active_nodes, self._inactive_nodes)
        self.config_lock.release()
        return ret

    def getPrimaryNode(self):
        self.config_lock.acquire()
        self.check_reload()
        ret = self._primary_node
        self.config_lock.release()
        return ret

    def getFailoverTimeoutSeconds(self) -> int:
        self.config_lock.acquire()
        self.check_reload()
        ret = self._failover_sampling_interval
        self.config_lock.release()
        return ret

    def check_reload(self):
        """Check config reload.

        Returns True if reload happened, False otherwise.
        """
        if self.last_mtime != getmtime(self.config_file):
            self.load_config()
            return True
        return False

    def who_am_I(self):
        self.config_lock.acquire()
        self.check_reload()
        ret = self._my_name
        self.config_lock.release()
        return ret

    def load_config(self):
        try:
            node_config = NodeConfig()
            root = node_config.get_current_config_root(self.config_file)
            last_mtime = getmtime(self.config_file)
        except Exception:
            self.logger.warning(
                f'Failed to parse config file {self.config_file}.',
                exc_info=True
            )
            return False

        node_tmp = root.findall('./DesiredNodes/Node')
        if len(node_tmp) == 0:
            self.logger.warning(
                f'The config file {self.config_file} is missing entries '
                'in the DesiredNodes section'
            )
            return False

        desired_nodes = [node.text for node in node_tmp]
        active_nodes = [
            node.text for node in root.findall('./ActiveNodes/Node')
        ]
        inactive_nodes = [
            node.text for node in root.findall('./InactiveNodes/Node')
        ]

        node_tmp = root.find('./PrimaryNode')
        if node_tmp is None or len(node_tmp.text) == 0:
            self.logger.warning(
                f'The config file {self.config_file} is missing a valid '
                'PrimaryNode entry'
            )
            return False
        primary_node = node_tmp.text

        # find my name in this cluster
        names = set(node_config.get_network_addresses_and_names())
        all_nodes = set(desired_nodes)
        intersection = all_nodes & names
        if len(intersection) > 1:
            my_name = intersection.pop()
            self.logger.warning(
                'This node has multiple names in the list of desired nodes, '
                'was it added more than once? Some things may not work in '
                f'this configuration. Using {my_name} as the name for this '
                'node.'
            )
        elif len(intersection) == 0:
            self.logger.warning(
                'This node has no entry in the list of desired nodes.'
            )
            my_name = None
        elif len(intersection) == 1:
            my_name = intersection.pop()
            # handles the initial 0-node special case
            if my_name == '127.0.0.1':
                my_name = None

        self.logger.info(f'Loaded the config file, my name is {my_name}')

        desired_nodes.sort()
        active_nodes.sort()
        inactive_nodes.sort()
        self._desired_nodes = desired_nodes
        self._active_nodes = active_nodes
        self._inactive_nodes = inactive_nodes
        self._primary_node = primary_node
        self.last_mtime = last_mtime
        self._my_name = my_name

        sampling_interval_node = root.find('./CMAPIConfig/FailoverSamplingIntervalSeconds')
        if sampling_interval_node is not None and sampling_interval_node.text is not None:
            self._failover_sampling_interval = int(sampling_interval_node.text)
        else:
            self._failover_sampling_interval = DEFAULT_FAILOVER_SAMPLING_INTERVAL_SECS

        return True



import logging
import socket


MAGIC_BYTES = 0x14fbc137.to_bytes(4, 'little')
# value is tuple(command_bytes, command_value_length)
DBRM_COMMAND_BYTES = {
    'readonly':           ((20).to_bytes(1, 'little'), 0),
    'set_readonly':       ((14).to_bytes(1, 'little'), 0),
    'set_readwrite':      ((15).to_bytes(1, 'little'), 0),
    'set_system_state':   ((55).to_bytes(1, 'little'), 4),
    'get_system_state':   ((54).to_bytes(1, 'little'), 4),
    'clear_system_state': ((57).to_bytes(1, 'little'), 4),
}
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 8616
SOCK_TIMEOUT = 5


class DBRMSocketHandler():
    """Class for stream socket operations.

    Include all logic for detecting bytestream protocol version, reading and
    parsing magic inside, getting command bytes and command value length
    by command name.

    """
    long_strings = None

    def __init__(
        self, family=socket.AF_INET, type=socket.SOCK_STREAM, proto=0,
        fileno=None
    ) -> None:
        self._socket = None
        self._family = family
        self._type = type
        self._proto = proto
        self._fileno = fileno
        self._host = None
        self._port = None
        self._recreate_socket()

    @property
    def _connect_called(self):
        """Is connect method called previously.

        This is the instance state to determine if "connect" method called
        previously. This is not quaranteed that connection still alive.
        :return: connected state
        :rtype: bool
        """
        if self._host and self._port:
            return True
        return False

    def _recreate_socket(self) -> None:
        """Create new internal _socket object.

        Create\recreate new _socket object and connects to it if was already
        connected.
        """
        if self._socket is not None:
            self._socket.close()
        self._socket = socket.socket(
            family=self._family, type=self._type,
            proto=self._proto, fileno=self._fileno
        )
        if self._connect_called:
            self.connect(self._host, self._port)

    def _detect_protocol(self) -> None:
        """Detect dbrm socket bytestream version.

        This method normally will be called only in first instance
        at first "send" method call.
        After that header will be formed and parsed depending on
        "long_strings" class variable value.

        Sends "readonly" message with "old" protocol version (before MCS 6.2.1)
        If timeout error raised, sends message with "new" protocol version
        (after MCS 6.2.1) with extra 4 bytes in header.
        If both attemts are failed raise RuntimeError and return the
        "long_strings" variable to initial state - None.

        :raises RuntimeError: [description]
        """
        success = False
        # check at first old protocol because 5.x.x version got an issue if
        # we try to send new format packages.
        for long_strings in (False, True):
            DBRMSocketHandler.long_strings = long_strings
            self.send('readonly')
            try:
                _ = self.receive()
                success = True
                break
            except (socket.timeout, TimeoutError):
                # wrong packet sended could cause errors on the mcs engine side
                self._recreate_socket()
                continue
        if not success:
            # something went wrong so return to unknown protocol state
            DBRMSocketHandler.long_strings = None
            raise RuntimeError(
                'Can\'t detect DBRM bytestream protocol version.'
            )
        else:
            dbrm_protocol_version = (
                'new' if DBRMSocketHandler.long_strings else 'old'
            )
            logging.info(
                f'Detected "{dbrm_protocol_version}" DBRM bytestream protocol'
            )

    def _make_msg(self, command_name: str, command_value: int) -> bytes:
        """Make bytes msg by command name and value.

        :param command_name: name of a command
        :type command_name: str
        :param command_value: command value
        :type command_value: int or None
        :return: msg to send throught socket
        :rtype: bytes
        """
        command_bytes, command_value_length = DBRM_COMMAND_BYTES[command_name]
        data_length = (
            command_value_length + len(command_bytes)
        ).to_bytes(4, 'little')
        # bytestream protocol before MCS 6.2.1 version
        package_header = MAGIC_BYTES + data_length
        if DBRMSocketHandler.long_strings:
            # bytestream protocol after MCS 6.2.1 version
            long_strings_count = (0).to_bytes(4, 'little')
            package_header += long_strings_count

        msg_bytes = package_header + command_bytes
        if command_value is not None:
            msg_bytes += command_value.to_bytes(
                command_value_length, 'little'
            )
        return msg_bytes

    def _receive_magic(self):
        """Reads the stream up to the uncompressed magic.

        The magic is a constant delimeter that occurs at the begging
        of the stream.
        """
        data: bytes
        recv_data: bytes = b''
        while recv_data != MAGIC_BYTES:
            data = self._socket.recv(1)
            # TODO: advanced error handling
            if data == b'':
                raise RuntimeError(
                    'Socket connection broken while receiving magic'
                )
            recv_data += data
            if not MAGIC_BYTES.startswith(recv_data):
                recv_data = data
                continue

    def _receive(self, length: int):
        """Receive raw data from socket by length.

        :param length: length in bytes to receive
        :type length: int
        :raises RuntimeError: if socket connection is broken while receiving
        :return: received bytes
        :rtype: bytes
        """
        chunks = []
        bytes_recd = 0
        while bytes_recd < length:
            chunk = self._socket.recv(min(length - bytes_recd, 2048))
            if chunk == b'':
                raise RuntimeError(
                    'Socket connection broken while receiving data.'
                )
            chunks.append(chunk)
            bytes_recd += len(chunk)
        return b''.join(chunks)

    def _send(self, msg: bytes):
        """Send msg in bytes through the socket.

        :param msg: string in bytes to send
        :type msg: bytes
        :raises RuntimeError: if connection is broken while sending
        """
        totalsent = 0
        while totalsent < len(msg):
            sent = self._socket.send(msg[totalsent:])
            if sent == 0:
                raise RuntimeError(
                    'DBRM socket connection broken while sending.'
                )
            totalsent = totalsent + sent

    def connect(self, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT):
        """Connect to socket.

        By default it connects with DBRM master.
        """
        self._host = host
        self._port = port
        self._socket.settimeout(SOCK_TIMEOUT)
        self._socket.connect((host, port))

    def close(self):
        """Closing the socket.

        Set _host and _port instance variables to None to change state to
        not connected. Then close the _socket.
        """
        self._host = None
        self._port = None
        self._socket.close()

    def send(self, command_name: str, command_value: int = None):
        """Top level send by command name and value.

        param command_name: name of a command
        :type command_name: str
        :param command_value: command value, defaults to None
        :type command_value: int, optional
        """
        if DBRMSocketHandler.long_strings is None:
            self._detect_protocol()
        msg_bytes = self._make_msg(command_name, command_value)
        self._send(msg_bytes)

    def receive(self):
        """Top level method to receive data from socket.

        Automatically reads the magic and data length from data header.

        :return: received bytes without header
        :rtype: bytes
        """
        self._receive_magic()
        data_length = int.from_bytes(self._receive(4), 'little')
        if DBRMSocketHandler.long_strings:
            # receive long strings count to meet new bytestream protocol
            # requirements (after MCS 6.2.1 release)
            long_strings_count_bytes = self._receive(4)
        data_bytes = self._receive(data_length)
        return data_bytes
